#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import copy
from typing import Dict

# PyQt5 Imports
from PyQt5.QtCore import QSizeF, QPointF

# Project Imports
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.core.plci_core_data_type import get_data_type
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.interface.files.abstract_importer import AbstractImporter
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.links_model.link_type import get_link_type
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel

#-----------------------------------
# Functions
#-----------------------------------

def construct_box (box_dict : Dict) -> BoxModel :
    """Est la méthode appelée pour produire un modèle de box suivant un dictionnaire d'import."""

    # Get genral box informations
    library = box_dict['library']
    name = box_dict['name']

    #
    text = box_dict['text']

    # Get geometry
    position = QPointF(float(box_dict['geometry']['x']), float(box_dict['geometry']['y']))
    size = QSizeF(float(box_dict['geometry']['width']), float(box_dict['geometry']['height']))
    rotation = float(box_dict['geometry']['rotation'])

    # Build Box from BoxData
    box_data = BoxesLibrariesManager().get_box_data(library, name)
    
    # Handle Geometry value
    ## Size
    if size.width() < AbstractBoxModel.MinimunBoxWidth:
        size.setWidth(AbstractBoxModel.MinimunBoxWidth)
    if size.height() < AbstractBoxModel.MinimunBoxHeight:
        size.setHeight(AbstractBoxModel.MinimunBoxHeight)
    
    ## Rotation
    rotation = ((rotation // 90) % 4) * 90

    # Build Box Model
    box_model = BoxModel(
        library=box_data.box_library,
        name=box_data.box_name,
        position=position,
        size=size,
        rotation=rotation, can_be_loaded=True, text=text, icon=box_data.icon
    )
    
    # Add default params value
    if hasattr(box_data, 'params') :
        for parameter, attributes in box_data.params.items() :
            # Copy parameter
            box_model._parameters[parameter] = copy.deepcopy(attributes)

    return box_model


def construct_input (input_dict : Dict, current_input_index : int, box_model : BoxModel) -> InputConnectionModel :
    """Est la méthode appelée pour produire un modèle d'entrée d'une box suivant un dictionnaire d'import."""

    input_name = input_dict['name']
    input_type = get_data_type(input_dict['data_type'])
    input_text = input_dict['text']
    input_is_infinite = input_dict['isInfinite']

    return box_model.insert_input(current_input_index, input_name)

    return box_model.insert_input(current_input_index, input_name, input_type, input_text, input_is_infinite)

def construct_output (output_dict : Dict, current_output_index : int, box_model : BoxModel) -> OutputConnectionModel : 
    """Est la méthode appelée pour produire un modèle de sortie d'une box suivant un dictionnaire d'import."""

    output_name = output_dict['name']
    output_type = get_data_type(output_dict['data_type'])
    output_text = output_dict['text']
    output_is_infinite = output_dict['isInfinite']

    return box_model.insert_output(current_output_index, output_name)
    # return box_model.insert_output(current_output_index, output_name, output_type, output_text, output_is_infinite)

#-----------------------------------
# Class
#-----------------------------------

class SimulationsImporter (AbstractImporter) :
    """Est la classe qui permet d'importer une simulation."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        
        AbstractImporter.__init__(self)

    # -------------
    # Methods
    # -------------

    def open (self, simulation_data : Dict, path : str, name : str) -> SimulationModel :
    
        if simulation_data['export_version'] == '1.0.0' :
            return self.v_1_0_0(simulation_data, path, name)
        elif simulation_data['export_version'] == '1.1.0' :
            return self.v_1_1_0(simulation_data, path, name)

        return None

    def v_1_0_0 (self, simulation_data : Dict, path : str, name : str) -> SimulationModel : 
        
        # General Project informations
        date_creation = simulation_data['date_creation']
        version = float(simulation_data['version'])
        creator = simulation_data['creator']

        # Simulation informations
        stop_time = float(simulation_data['stop_time'])
        step_time = float(simulation_data['step_time'])
        scheduler_library = simulation_data['scheduler']['library']
        scheduler_name = simulation_data['scheduler']['name']

        # Build Scehduler & Simulation Models
        scheduler_params = SchedulerParams(stop_time, step_time)
        scheduler_model = SchedulerModel(scheduler_library, scheduler_name, scheduler_params)
        simulation_model = SimulationModel(name, path, creator, date_creation, version, scheduler_model)

        # Get list of all boxes & signals
        boxes = simulation_data['diagram']['boxes']
        signals = simulation_data['diagram']['signals']

        #Ajout de input et ouput au dictionnaire importé dans le but de connecter correctement les noeuds.
        #Plus rapide que de copier l'éxistant et d'ajotuer les "input" et "output". 
        
        for signal in signals : 
            signal['input'] = None
            signal['output'] = None

        for box_data in boxes :

            box_model = construct_box(box_data)
            
            current_input_index = 0
            for input_data in box_data['inputs'] : 

                input = construct_input(input_data, current_input_index, box_model)
                
                input_signal_index = int(input_data['signal_index'])
                if input_signal_index < len(signals) and input_signal_index >= 0 : 

                    signal_data = signals[input_signal_index]
                    if signal_data['input'] is None : 
                        signal_data['input'] = input
                        if not(signal_data['output'] is None) : 
                            simulation_model.link_nodes(input, signal_data['output'], -1, -1,
                                float(signal_data['link_value']), get_link_type(signal_data['link_type']), '')

                current_input_index += 1

            current_output_index = 0
            for output_data in box_data['outputs'] :

                output = construct_output(output_data, current_output_index, box_model)
                
                for output_signal_index in output_data['signal_indices'] : 
                    if output_signal_index < len(signals) and output_signal_index >= 0 : 
                        
                        signal_data = signals[output_signal_index]
                        if signal_data['output'] is None : 
                            signal_data['output'] = output
                            if not(signal_data['input'] is None) : 
                                simulation_model.link_nodes(signal_data['input'], output, -1, -1,
                                    float(signal_data['link_value']), get_link_type(signal_data['link_type']), '')
                
                current_output_index += 1


            # Set parameters
            for parameter_data in box_data['params'] :
                box_model.set_parameter(parameter_data, box_data['params'][parameter_data]['value'])

            # Append Box Model
            simulation_model.add_element(box_model)

        return simulation_model

    def v_1_1_0 (self, simulation_data : Dict, path : str, name : str) -> SimulationModel : 
        
        # General Project informations
        date_creation = simulation_data['date_creation']
        version = float(simulation_data['version'])
        creator = simulation_data['creator']

        # Simulation informations
        stop_time = float(simulation_data['stop_time'])
        step_time = float(simulation_data['step_time'])
        scheduler_library = simulation_data['scheduler']['library']
        scheduler_name = simulation_data['scheduler']['name']

        # Build Scehduler & Simulation Models
        scheduler_params = SchedulerParams(stop_time, step_time)
        scheduler_model = SchedulerModel(scheduler_library, scheduler_name, scheduler_params)
        simulation_model = SimulationModel(name, path, creator, date_creation, version, scheduler_model)

        # Get list of all boxes & signals
        boxes = simulation_data['diagram']['boxes']
        signals = simulation_data['diagram']['signals']

        #Ajout de input et ouput au dictionnaire importé dans le but de connecter correctement les noeuds.
        #Plus rapide que de copier l'éxistant et d'ajotuer les "input" et "output". 
        
        for signal in signals : 
            signal['input'] = None
            signal['output'] = None

        for box_dict in boxes :

            # Get default box data from dict informations
            box_model = BoxModel.fromDict(box_dict)

            # Build infinite inputs groups
            for inputs_group in box_dict["inputs_groups"]:

                # Get group data
                is_infinite = inputs_group["isInfinite"]
                inputs_group_name = inputs_group["name"]

                if is_infinite:
                    # Add inputs
                    for input_data in inputs_group['inputs'] : 
                        # Input data
                        input_signal_index = int(input_data["signal_index"])
                        input_text = input_data["text"]

                        # Append input to box model
                        input_ = box_model.append_input(inputs_group_name)

                        # Set input parameter
                        input_.text = input_text

                        if input_signal_index < len(signals) and input_signal_index >= 0 : 
                            # Get signal data
                            signal_data = signals[input_signal_index]

                            # Add input to signal data
                            if signal_data['input'] is None : 
                                signal_data['input'] = input_

                else:
                    # Set inputs
                    for input_index, input_data in enumerate(inputs_group['inputs']): 
                        # Input data
                        input_signal_index = int(input_data["signal_index"])
                        input_text = input_data["text"]

                        # Get input from box model
                        input_ = box_model.inputs_groups[inputs_group_name].inputs[input_index]

                        # Set input parameter
                        input_.text = input_text

                        if input_signal_index < len(signals) and input_signal_index >= 0 : 
                            # Get signal data
                            signal_data = signals[input_signal_index]

                            # Add input to signal data
                            if signal_data['input'] is None : 
                                signal_data['input'] = input_

            
            # Build infinite outputs groups
            for outputs_group in box_dict["outputs_groups"]:

                # Get group data
                is_infinite = outputs_group["isInfinite"]
                outputs_group_name = outputs_group["name"]

                if is_infinite:
                    # Add outputs
                    for output_data in outputs_group['outputs'] : 
                        # Input data
                        output_signal_indices = output_data["signal_indices"]
                        output_text = output_data["text"]

                        # Append output to box model
                        output_ = box_model.append_output(outputs_group_name)

                        # Set output parameter
                        output_.text = output_text

                        for output_signal_index in output_signal_indices:

                            if output_signal_index < len(signals) and output_signal_index >= 0 : 
                                # Get signal data
                                signal_data = signals[output_signal_index]

                                # Add output to signal data
                                if signal_data['output'] is None : 
                                    signal_data['output'] = output_

                else:
                    # Set outputs
                    for output_index, output_data in enumerate(outputs_group['outputs']): 
                        # Input data
                        output_signal_indices = output_data["signal_indices"]
                        output_text = output_data["text"]

                        # Get output from box model
                        output_ = box_model.outputs_groups[outputs_group_name].outputs[output_index]

                        # Set output parameter
                        output_.text = output_text

                        for output_signal_index in output_signal_indices:
                            if output_signal_index >= 0 : 
                                # Get signal data
                                signal_data = signals[output_signal_index]

                                # Add output to signal data
                                if signal_data['output'] is None :
                                    signal_data['output'] = output_

            # Set parameters
            for parameter_data in box_dict['params'] :
                box_model.set_parameter(parameter_data, box_dict['params'][parameter_data])

            # Append Box Model
            simulation_model.add_element(box_model)

            
        # Signals
        for signal_data in signals:
            # Get signal data
            input_ = signal_data["input"]
            output = signal_data["output"]
            
            # Create link
            if input_ is not None and output is not None:
                simulation_model.link_nodes(input_, output, -1, -1,
                    float(signal_data['link_value']), get_link_type(signal_data['link_type']), '')

        return simulation_model


#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtCore import QPointF, QSizeF

from PyQt5.QtWidgets import QGraphicsItem
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel


#-----------------------------------
# Class
#-----------------------------------

class AbstractBoxIOModel (AbstractBoxModel) :
    """Est la classe des entrées des boxes. Elles peuvent-être drag and drop dans l'interface."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, size : QSizeF, rotation : float = 0.0,
                  text : str = '', parent : QGraphicsItem = None) :
        
        AbstractBoxModel.__init__(self, name, position, size, rotation, text, parent)

        self._data_type : DataType = data_type
    
    # -------------
    # Properties
    # -------------

    @property
    def data_type (self) -> DataType :
        """Permet de récuperer le type de lien."""
        return self._data_type

    @data_type.setter
    def data_type (self, data_type_ : DataType) -> None :
        """Permet de modifier le type du lien."""
        self._data_type = data_type_

    # -------------
    # Properties
    # -------------

    def insert_input (self, index : int, input : InputConnectionModel) -> None :
        """Permet d'insérer une entrée dans la box."""
        ...

    def remove_input (self, input : InputConnectionModel) -> None :
        """Permet de suppprimer une entrée de la box."""
        ...

    def insert_output (self, index : int, output : OutputConnectionModel) -> None :
        """Permet d'insérer une sortie dans la box."""
        ...

    def remove_output (self, output : OutputConnectionModel) -> None : 
        """Permet de supprimer une sortie de la box."""
        ...


import os
from threading import Lock
from muphyn.packages.core.utils.singleton import SingletonMetaClass

# class DirectoryManagerMeta(type):
#     """
#     This is a thread-safe implementation of Singleton.
#     """

#     _instances = {}

#     _lock: Lock = Lock()
#     """
#     We now have a lock object that will be used to synchronize threads during
#     first access to the Singleton.
#     """

#     def __call__(cls, *args, **kwargs):
#         with cls._lock:
#             if cls not in cls._instances:
#                 instance = super().__call__(*args, **kwargs)
#                 cls._instances[cls] = instance
#         return cls._instances[cls]

class DirectoryManager(metaclass=SingletonMetaClass):

    def __init__(self) -> None:
        # Init base working directory
        self.base_working_directory = os.getcwd()
        
        # Init current working directory
        self.current_working_directory = self.base_working_directory

    def reset_working_directory(self):
        # Test if Base Working Directory exists
        if not os.path.exists(self.base_working_directory):
            raise Exception(f"Base Working Directory doesn't exists: {self.base_working_directory}")
            
        # Change working directory
        self.set_working_directory(self.base_working_directory)

    def set_working_directory(self, new_wd_path: str):
        # Test IF new path exists & path is dir
        if os.path.exists(new_wd_path) and os.path.isdir(new_wd_path):
            # Test IF new path is relative
            if not os.path.isabs(new_wd_path):
                # Build absolute path from relative path
                new_wd_path = os.path.join(self.current_working_directory, new_wd_path)

            # Change working directory
            os.chdir(new_wd_path)
            self.current_working_directory = os.getcwd().replace("\\", "/")
        else:
            raise Exception(f"New Working Directory doesn't exists : {new_wd_path}")



#-----------------------------------
# Imports
#-----------------------------------

from signal import signal
from typing import Iterable, List

from PyQt5 import QtGui
from PyQt5.QtCore import QMimeData, QPointF, QSizeF, Qt
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsSceneMouseEvent, QStyleOptionGraphicsItem

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.models.signals_model.abstract_connection_model import AbstractConnectionModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.models.links_model.temporary_link_model import TemporaryLink

#-----------------------------------
# Fnctions
#-----------------------------------

def start_drag (signal_creator : AbstractConnectionModel, event: QGraphicsSceneMouseEvent) -> bool :
    mimeData : QMimeData = QMimeData()
    mimeData.setData('action', bytearray('new link'.encode()))
    mimeData.setData('link', bytearray(str(id(signal_creator)).encode()))
    mimeData.setData('type', bytearray(signal_creator.data_type.__str__().encode()))

    drag : QtGui.QDrag = QtGui.QDrag(signal_creator)
    drag.setMimeData(mimeData)
    da : Qt.DropAction = drag.exec(Qt.LinkAction)

    # del temporary_link

#-----------------------------------
# Class
#-----------------------------------

class SignalNodeModel (AbstractConnectionModel) :
    """Est le modèle des noeuds permettant de diviser les liens en plusieurs segments."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, size : QSizeF, links : Iterable[SignalLinkModel] = [],
                  text : str = '', is_infinite : bool = False, parent : QGraphicsItem = None) :

        AbstractConnectionModel.__init__(self, name, data_type, position, size, links, text, is_infinite, parent)
        
    # -------------
    # Properties
    # -------------

    @property 
    def is_input (self) -> bool :
        """Permet de savoir si l'élément est une entrée."""
        return False

    @property
    def is_connected_to_input (self) -> bool :
        """Permet de savoir si l'élément actuel est connecté à une entrée (ou est un entrée)."""

        if self.is_input :
            return True
        
        is_connected = False

        lst_already_passed : List[AbstractConnectionModel] = [self]
        lst_to_pass : List[AbstractConnectionModel] = [] 

        for link in self._links :

            if not(link.node_1 in lst_already_passed) :
                if not(link.node_1 in lst_to_pass) :
                    lst_to_pass.append(link.node_1)
            
            if not(link.node_2 in lst_already_passed) :
                if not(link.node_2 in lst_to_pass) :
                    lst_to_pass.append(link.node_2)

        while lst_to_pass.__len__() > 0 :
            
            connectionModel : AbstractConnectionModel = lst_to_pass[0]
            
            if connectionModel.is_input :
                return True
            
            for link in connectionModel._links :

                if not(link.node_1 in lst_already_passed) :
                    if not(link.node_1 in lst_to_pass) :
                        lst_to_pass.append(link.node_1)
                
                if not(link.node_2 in lst_already_passed) :
                    if not(link.node_2 in lst_to_pass) :
                        lst_to_pass.append(link.node_2)

            lst_to_pass.remove(connectionModel)
            lst_already_passed.append(connectionModel)
        
        return is_connected

    # -------------
    # Methods
    # -------------
    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None :

        if (event.button() == Qt.LeftButton and event.modifiers() == Qt.ControlModifier) :
            start_drag(self, event)
            event.accept() 

        return super().mousePressEvent(event)



from .unit_test import get_boxes_libraries

# Get library
boxesLibraries = get_boxes_libraries()

print("param for graph box : ", boxesLibraries.get_required_params('Boxes.Sinks', 'Graph'))

print("param for wrong box : ", boxesLibraries.get_required_params('Boxes.Sinks', 'E'))

print("param for wrong library : ", boxesLibraries.get_required_params('Boxes.Sinks', 'Ramp'))

print("param for error in library : ", boxesLibraries.get_required_params(1, 'Ramp'))


from muphyn.packages.unit_test.unit_test import get_schedulers_library

schedulersLibraries = get_schedulers_library()

print("scheduler loaded : ", schedulersLibraries.libraries[0].schedulers.__len__())


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

from muphyn.packages.core.utils.log_manager import LogManager

#-----------------------------------
# Functions
#-----------------------------------

def _init_square_box_2 (box, simulation_params) -> None :
    
    box['current_period'] = 0

    if not 'high_value' in box :
        box['high_value'] = 1.0

    if not 'low_value' in box :
        box['low_value'] = 0.0

    if not 'high_time' in box :
        box['high_time'] = 0.5

    if not 'low_time' in box :
        box['low_time'] = 0.5

    if not 'rise_time' in box: 
        box['rise_time'] = 0.01

    if not 'fall_time' in box: 
        box['fall_time'] = 0.01

    if box['fall_time'] == 0 :
        box['fall_slope'] = 1
    else :
        box['fall_slope'] = (box['low_value'] - box['high_value']) / box['fall_time']

    if box['rise_time'] == 0:
        box['rise_slope'] = 1
    else :
        box['rise_slope'] = (box['high_value'] - box['low_value'])  / box['rise_time']

    LogManager().debug("Creating square box")
    LogManager().debug(f'\t - high_time : {box["high_time"]}')
    LogManager().debug(f'\t - low_time : {box["low_time"]}')
    LogManager().debug(f'\t - rise_time : {box["rise_time"]}')
    LogManager().debug(f'\t - fall_time : {box["fall_time"]}')
    LogManager().debug(f'\t - period : {box["high_time"] + box["low_time"]}')
    LogManager().debug(f'\t - rise slope : {box["rise_slope"]}')
    LogManager().debug(f'\t - fall slope : {box["fall_slope"]}')

    LogManager().debug('-- computing --')
    box['high_time'] = box['high_time'] - box['fall_time']
    LogManager().debug(f'\t - high_time \t= high_time - fall_time = {box["high_time"]}')
    box['fall_time'] = box['high_time'] + box['fall_time'] 
    LogManager().debug(f'\t - fall_time \t= high_time + fall_time = {box["fall_time"]}')
    box['low_time'] = box['low_time'] + box['fall_time'] - box['rise_time']
    LogManager().debug(f'\t - low_time \t= low_time + fall_time - rise_time = {box["low_time"]}')
    box['rise_time'] = box['low_time'] + box['rise_time']
    LogManager().debug(f'\t - rise_time \t= low_time + rise_time = {box["rise_time"]}')


def _function_square_box_2 (box, event_) -> List : 

    if box['current_period'] < box['high_time'] :

        v = box['high_value']

    elif box['current_period'] < box['fall_time']:

        v = box['high_value'] + (box['fall_slope'] * (box['fall_time'] - box['current_period']))

    elif box['current_period'] < box['low_time']:
        
        v = box['low_value']

    elif box['current_period'] < box['rise_time']:

        v = box['low_value'] + (box['rise_slope'] * (box['rise_time'] - box['current_period']))
    
    else :

        v = box['high_value']

    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))
    
    box['current_period'] = box['current_period'] + event_.step_time

    while box['current_period'] >= box['rise_time'] :
        box['current_period'] = box['current_period'] - box['rise_time']

    return events


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from typing import Iterable, List, Any

# Project Imports
from muphyn.packages.interface.graphical_actions.abstract_diagram_action import AbstractDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.graphical_actions import parser_decode as parser_decode
from muphyn.packages.interface.graphical_actions import parser_encode as parser_encode

#-----------------------------------
# Class
#-----------------------------------

class DiagramRemoveGraphicalElementAction (AbstractDiagramAction) :
    """Est l'action capable de supprimer la sélection actuelle de box."""

    # -------------
    # Methods
    # -------------
    
    def __init__ (self, diagram_model : Any, elements : Iterable[Any]) :

        AbstractDiagramAction.__init__(self, diagram_model)
        
        self._elements_index = []
        for el in elements :
            if isinstance(el, AbstractBoxModel) or isinstance(el, AbstractLinkModel) : 
                self._elements_index.append(el.graphical_index)

        self._reconstructors = []

    # -------------
    # Methods
    # -------------
    
    def elements (self) -> Iterable[AbstractGraphicalElement] :
        """Permet de récuperer les éléments à supprimer directement depuis leur indexs."""

        for index in self._elements_index :

            el = self.diagram_model.get_element_by_graphical_index(index)
            if not(el is None) : 
                yield el

    def do (self) :
        
        to_delete : List[AbstractBoxModel] = []
        for box_model in self.elements() :
            if isinstance(box_model, AbstractBoxModel) :
                to_delete.append(box_model)

        for link in self.elements() :
            if isinstance(link, AbstractLinkModel) : 
                self._reconstructors.append(parser_encode.link(link))
                link.unbind()
                self.diagram_model.remove_element(link)
                
        for box in to_delete :
            signals = list(box.signals)
            for index, link in enumerate(signals):
                self._reconstructors.append(parser_encode.link(link))
                link.unbind()
                self.diagram_model.remove_element(link)
        
        for box in to_delete :
            self._reconstructors.append(parser_encode.encode(box))
            self.diagram_model.remove_element(box)

        self._elements_index.clear()

    def undo (self) :
        
        # Rebuild all items except link
        for reconstructor_box in self._reconstructors :
            if not(reconstructor_box['type'] == 'link') :
                box_model = parser_decode.box(reconstructor_box)
                
                self.diagram_model.add_element(box_model)
                self._elements_index.append(box_model.graphical_index)

                box_model.setSelected(True)
                
        # Rebuild all links
        for reconstructor_link in self._reconstructors :
            if reconstructor_link['type'] == 'link' :

                link_model = parser_decode.link(reconstructor_link, self.diagram_model)
                
                self._elements_index.append(link_model.graphical_index)
                link_model.setSelected(True)

        self._reconstructors.clear()


#-----------------------------------
# Imports
#-----------------------------------

import os
import yaml
from typing import List, Any, Dict

from muphyn.packages.interface.models.links_model.link_type import LinkType, get_link_type

#-----------------------------------
# Class
#-----------------------------------

class UserData :
    """Sont les données utilisateurs."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, path : str) :
        
        self._path : str = path
        self._boxes_libraries : List[str] = []
        self._schedulers_libraries : List[str] = []
        self._user_name : str = 'No name'
        self._recent_files : List[str] = []
        self._lang : str = 'french'
        self._theme : str = 'light'
        self._grid_type : str = 'dots'
        self._attach_box_to_grid : bool = False
        self._grid_spacing : int = 5
        self._default_link_type : LinkType = LinkType.SQUARE
        self._default_link_curving : float = 5

    # -------------
    # Properties
    # -------------

    @property
    def boxes_libraries (self) -> List[str] :
        """Permet de récuperer la liste des bibliothèques de boxes que l'utilisateur souhaite utiliser."""
        return self._boxes_libraries

    @property
    def schedulers_libraries (self) -> List[str] :
        """Permet de récuperer la liste des bibliothèques de planificateurs que l'utilisateur souhaite utiliser."""
        return self._schedulers_libraries

    @property
    def user_name (self) -> str :
        """Permet de récuperer le nom d'utilisateur."""
        return self._user_name

    @user_name.setter
    def user_name (self, new_name : str) -> None :
        """Permet de modifier le nom d'utilisateur."""
        if new_name is None : 
            return

        if new_name == '' :
            return

        self._user_name : str = new_name

    @property
    def path (self) -> str :
        """Permet de récuperer le chemin vers le fichier dans lequel il faut lire et sauvegarder les données utilisateur."""
        return self._path

    @path.setter
    def file (self, path_ : str) -> None :
        """Permet de modifier le chemin vers le fichier dans lequel il faut lire et sauvegarder les données utilisateur."""
        if path_ is None : 
            return

        if path_ == '' :
            return

        self._path : str = path_
        

    @property
    def recent_files (self) -> List[str] :
        """Permet de récuperer les fichiers récement ouverts."""
        return self._recent_files

    @property
    def language (self) -> str : 
        """Permet de récuperer le language dans lequel l'utilisateur veut voir l'interface."""
        return self._lang

    @language.setter
    def language (self, lang_ : str) -> None :
        """Permet de modifier la langue dans laquelle l'utilisateur veut voir l'interface."""
        if lang_ is None:
            return

        if lang_ == '':
            return

        self._lang = lang_
        
    @property
    def theme (self) -> str:
        """Permet de récuperer le thème dans lequel l'utilisateur veut voir l'interface."""
        return self._theme

    @theme.setter
    def theme (self, theme_ : str) -> None :
        """Permet de modifier le thème dans lequel l'utilisateur veut voir l'interface."""
        if theme_ is None:
            return

        if theme_ == '':
            return

        self._theme = theme_

    @property
    def grid_type (self) -> str:
        """Permet de récuperer le type dans lequel il faut que la grille de l'éditeur s'affiche."""
        return self._grid_type

    @grid_type.setter
    def grid_type (self, grid_type_ : str) -> None :
        """Permet de modifier le type dans lequel il faut que la grille de l'éditeur s'affiche."""
        if grid_type_ is None:
            return

        if grid_type_ == '':
            return

        self._grid_type = grid_type_

    @property
    def attach_box_to_grid (self) -> bool :
        """Permet de récuperer la variable décrivant si les boxes sont attachées à la grille ou pas."""
        return self._attach_box_to_grid

    @attach_box_to_grid.setter
    def attach_box_to_grid (self, attach_box_to_grid_ : bool) -> None :
        """Permet de modifier la variable décrivant si les boxes sont attachées à la grille ou pas."""
        self._attach_box_to_grid : bool = attach_box_to_grid_

    @property
    def grid_spacing (self) -> int :
        """Permet de récuperer l'espacement de la grille."""
        return self._grid_spacing
    
    @grid_spacing.setter 
    def grid_spacing (self, grid_spacing_ : int) -> None :
        """Permet de modifier l'espacement de la grille."""
        if grid_spacing_ < 1 :
            return 

        self._grid_spacing : int = grid_spacing_

    @property
    def default_link_type (self) -> LinkType :
        """Permet de récuperer le type de liens entre les boxes par défaut."""
        return self._default_link_type

    @default_link_type.setter 
    def default_link_type (self, default_link_type_ : LinkType) -> None :
        """Permet de modifier le type de liens entre les boxes par défaut."""

        if default_link_type_ is None:
            return

        self._default_link_type = default_link_type_

    @property
    def default_link_curving (self) -> float :
        """Permet de récuperer la courbure par défaut des liens courbes."""
        return self._default_link_curving

    @default_link_curving.setter
    def default_link_curving (self, default_link_curving_ : float) -> None :
        """Permet de modifier la courbure par défaut des liens courbes."""
        if default_link_curving_ < 0 :
            return

        self._default_link_curving : float = default_link_curving_

    # -------------
    # Methods
    # -------------

    def save (self) :
        """Est la méthode pour sauvegarder les données utilisateur."""

        to_save : Dict[str, Any] = {}
        to_save['version'] = 1.0
        to_save['user_name'] = self._user_name
        to_save['boxes_libraries'] = self._boxes_libraries
        to_save['schedulers_libraries'] = self._schedulers_libraries
        to_save['recent_files'] = self._recent_files
        to_save['lang'] = self._lang
        to_save['theme'] = self._theme
        to_save['grid_type'] = self._grid_type
        to_save['attach_box_to_grid'] = self._attach_box_to_grid
        to_save['grid_spacing'] = self._grid_spacing
        to_save['default_link_type'] = self._default_link_type.__str__()
        to_save['default_link_curving'] = self._default_link_curving

        with open(self._path, 'w') as file:
            yaml.dump(to_save, file)

    def load (self) :
        """Est la méthode pour charger les données utilisateur."""

        if os.path.isfile(self._path) :
            
            with open(self._path, 'r') as file:
                document = yaml.full_load(file)

                if document['version'] == 1.0 :
                    self._user_name = document['user_name']
                    self._boxes_libraries = document['boxes_libraries']
                    self._schedulers_libraries = document['schedulers_libraries']
                    self._recent_files = document['recent_files']
                    self._lang = document['lang']
                    self._theme = document['theme']
                    self._grid_type = document['grid_type']
                    self._attach_box_to_grid = document['attach_box_to_grid']
                    self._grid_spacing = document['grid_spacing']
                    self._default_link_type = get_link_type(document['default_link_type'])
                    self._default_link_curving = document['default_link_curving']

        


#-----------------------------------
# Imports
#-----------------------------------
import os
from typing import List
from muphyn.packages.core.plci_core_box import Box
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.core.simulation.om_sim import OpenModelicaSimulation
from muphyn.packages.core.simulation.model_parser.open_modelica_model_parser import OpenModelicaModelParser
from muphyn.packages.core.plci_core_scheduler_event import SchedulerEvent


"""
Notes:
 - If the initialization of the node is made in the _init_function then we won't access to the inputs first value 
    → (must implement a way to create firsts values)
 - If the outputs are read before the step is made we get one  step lag ;
 - If the inputs are set before the step then we get one step lag
"""

#-----------------------------------
# Functions
#-----------------------------------
def _on_model_file_selected(box_model: BoxModel) -> List:
    # Get path value
    path_parameter = box_model.get_parameter("model_path")
    path = path_parameter["value"]

    # Init Simulation object 
    om_parser = OpenModelicaModelParser(path)

    # Set inputs
    inputs = om_parser.inputs
    for input_index, input_ in enumerate(inputs):
        if input_index >= box_model.input_len:
            box_model.append_input(input_["name"], DataType.FLOAT)
        else:
            box_model_input = box_model.get_input(input_index)
            box_model_input.name = input_["name"]
            box_model_input.data_type = DataType.FLOAT

    # Set outputs
    outputs = om_parser.outputs
    for output_index, output in enumerate(outputs):
        if output_index >= box_model.output_len:
            box_model.append_output(output["name"], DataType.FLOAT)
        else:
            box_model_output = box_model.get_output(input_index)
            box_model_output.name = output["name"]
            box_model_output.data_type = DataType.FLOAT

    return box_model

def _init_simulation (box, simulation_params: SchedulerParams) -> List :
    box.simulation_params = simulation_params
    return box

def _step_simumation (box: Box, event_: SchedulerEvent) -> List :
    # if first step → run first simulation
    if event_.timing == 0:
        # Get path
        path = box.get_parameter("model_path")

        # Get simulation parameters
        simulation_params: SchedulerParams = box.simulation_params

        # Model path parser
        om_parser = OpenModelicaModelParser(path)

        # Set all input values in model file
        for input_signal in box.inputs:
            om_parser.set_input_value(input_signal.input_name, input_signal.value)

        # Get new model file path
        tmp_dirname = os.path.join(os.path.dirname(path), "tmp")
        basename = os.path.basename(path)
        new_model_file_path = os.path.join(tmp_dirname, basename)
        
        # Save file as
        om_parser.save_file_as(new_model_file_path)

        # Init Simulation object 
        om_simulation = OpenModelicaSimulation(new_model_file_path)

        # Init Simulation
        start_time = -simulation_params.step_time # Starting one step before
        stop_time = simulation_params.stop_time + simulation_params.step_time # go one step further
        step_size = simulation_params.step_time
        om_simulation.set_simulation_options(
            start_time=start_time,
            stop_time=stop_time, 
            step_size=step_size
        )

        # Init Simulation
        om_simulation.init_simulation()

        # Run Simulation
        om_simulation.run_simulation()

        om_simulation.make_step()

        # Set om_system attribute
        box.om_simulation = om_simulation

    else:
        # Set inputs value
        muphyn_current_value = None
        for input_signal in box.inputs:
            box.om_simulation.set_node_value(input_signal.input_name, input_signal.value)
            if input_signal.input_name == "instant_current_value":
                muphyn_current_value = input_signal.value
                
        for input_signal in box.inputs:
            box.om_simulation.set_node_value(input_signal.input_name, input_signal.value)
            if input_signal.input_name == "instant_current_value":
                muphyn_current_value = input_signal.value

        before_step_input_value = box.om_simulation.get_node_value(input_signal.input_name)

        # Make steps
        box.om_simulation.make_step()

        after_step_input_value = box.om_simulation.get_node_value(input_signal.input_name)


    # Get output values
    events : List = []
    for output_signal in box.outputs:
        output_value = box.om_simulation.get_node_value(output_signal.output_name)
        events.append(box.construct_signal_event(output_signal, output_value))

    # Return outputs
    return events



def _finish_simulation (box) -> List :
    
    # Make steps
    box.om_simulation.make_step(is_last_step = True)

    v = 1
    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events



# PyQt5 imports
from itertools import count
from PyQt5.QtCore import QPointF, QRectF, QSizeF, QSize
from PyQt5.QtWidgets import QGraphicsObject

# Project imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.shapes.basic_shape import GroupedShapes

class VerticalLayout(GroupedShapes):
    def __init__(self, group_position: QPointF = QPointF(0, 0), parent = None) -> None:
        super().__init__(group_position, parent)

        # Init child items
        self.childConfigs: dict[QGraphicsObject, dict] = {}

        # Maximum size
        self._maximumWidth: int = 0
        self._maximumHeight: int = 0

        # Spacing
        self._spacing: int = 0
    
    # -------------
    # Properties
    # -------------
    @property
    def count(self) -> int:
        return len(list(self.childConfigs))

    @property
    def maximumWidth(self) -> int:
        return self._maximumWidth

    @maximumWidth.setter
    def maximumWidth(self, newMaximumWidth: int) -> None:
        if newMaximumWidth != self._maximumWidth:
            self.setMaximumWidth(newMaximumWidth)

    @property
    def maximumHeight(self) -> int:
        return self._maximumHeight

    @maximumHeight.setter
    def maximumHeight(self, newMaximumHeight: int) -> None:
        if newMaximumHeight != self._maximumHeight:
            self.setMaximumHeight(newMaximumHeight)

    @property
    def spacing(self) -> int:
        return self._spacing

    @spacing.setter
    def spacing(self, newSpacing: int) -> None:
        if newSpacing != self._spacing:
            self.setSpacing(newSpacing)

    # -------------
    # Methods
    # -------------
    def boundingRect(self) -> QRectF:
        return QRectF(QPointF(), QSizeF(self.maximumWidth, self.maximumHeight))

    def addItem(self, newItem: QGraphicsObject, height: int = 0):
        # Create config parameters dictionnary
        self.childConfigs[newItem] = {
            "height": height if height > 0 else 0
        }

        # Set parent item
        newItem.setParentItem(self)

        # Handle children positions
        self.recomputeChildrenPosition()

    def removeItem(self, item: QGraphicsObject):
        # Remove item from child list
        del self.childConfigs[item]

        # Remove parent Item
        item.setParentItem(None)

        # Handle children positions
        self.recomputeChildrenPosition()

    def recomputeChildrenPosition(self):
        heightNotSetItems = [item for item, config in self.childConfigs.items() if config["height"] < 1]

        # Calculate total height
        totalMinimumSize = sum([config["height"] for config in self.childConfigs.values()])

        # Calculate total spacing height
        totalSpacingHeight = self._spacing * (self.count - 1) 

        # Calculate exceed height space
        exceedHeightSpace = self._maximumHeight - totalMinimumSize - totalSpacingHeight
        
        if exceedHeightSpace >= 0:
            # Calculate shared height space
            if len(heightNotSetItems) > 0:
                exceedHeightShared = exceedHeightSpace / len(heightNotSetItems)
            else:
                exceedHeightShared = 0

            # Place all items
            y_axis_position = 0
            for item, config in self.childConfigs.items():
                # Move Item
                item.setPos(QPointF(0, y_axis_position))

                # Set height
                if config["height"] == 0:
                    height = exceedHeightShared
                else:
                    height = config["height"]
                item.setSize(QSize(int(self.maximumWidth), int(height)))

                # Update next item y-axis position
                y_axis_position += height + self._spacing

        else:
            pass

    def setMaximumWidth(self, newMaximumWidth: int):
        if newMaximumWidth < 0:
            LogManager().error(f"VerticalLayout.setMaximumWidth(): Maximum width value is negative {newMaximumWidth}")
            return

        self._maximumWidth = newMaximumWidth

    def setMaximumHeight(self, newMaximumHeight: int):
        if newMaximumHeight < 0:
            LogManager().error(f"VerticalLayout.setMaximumHeight(): Maximum width value is negative {newMaximumHeight}")

        self._maximumHeight = newMaximumHeight

    def setSpacing(self, newSpacing: int):
        if newSpacing < 0:
            LogManager().error(f"VerticalLayout.setSpacing(): Spacing value is negative {newSpacing}")

        self._spacing = newSpacing



#-----------------------------------
# Imports
#-----------------------------------


import os
from datetime import date
from typing import List, Any

from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5.QtCore import QCoreApplication, QMetaObject, QPointF, QRect, QSizeF, pyqtSlot
from PyQt5.QtWidgets import QLabel, QLineEdit, QPushButton, QTableWidget, QWidget
from PyQt5.QtWidgets import QAbstractItemView, QComboBox, QFileDialog, QFrame

from muphyn.packages.core.plci_core_data_type import get_data_type
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.graphical_actions.new_box_dialog_add_action import NewBoxDialogAddAction
from muphyn.packages.interface.graphical_actions.new_box_dialog_remove_action import NewBoxDialogRemoveAction
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel
from muphyn.packages.interface.user_data.user_data import UserData
from .abstract_dialog import AbstractDialog

from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.editable_models.box_composite_model import BoxCompositeModel
from muphyn.packages.interface.models.editable_models.box_code_model import BoxCodeModel

#-----------------------------------
# Class
#-----------------------------------

class NewBoxDialog (AbstractDialog) : 
    """Est la classe permettant d'afficher une boîte de dialogue capable de créer une nouvelle box."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, dialog_holder : Any, user_data : UserData) :
        AbstractDialog.__init__(self, dialog_holder, 'new_simulation', 'Nouvelle Simulation')
        
        self.setFixedSize(630, 480)

        self._user_data = user_data
        self._actions_holder = ActionsHolder()

        self._init_ui()

        self._test_data_accept_button()
        self._tbl_inputs_selection_changed()
        self._tbl_outputs_selection_changed()

    # -------------
    # Properties
    # -------------

    @property
    def user_data (self) -> UserData :
        """Permet de récuperer les données utilisateurs."""
        return self._user_data
    
    # -------------
    # Methods
    # -------------

    def _init_ui (self) :
        """Est la méthode pour remplire l'interface graphique aves les widgets nécessaires."""
        if not self.objectName():
            self.setObjectName("_new_box_dialog")

        self._btn_cancel = QPushButton(self)
        self._btn_cancel.setObjectName("_btn_cancel")
        self._btn_cancel.setGeometry(QRect(540, 440, 81, 24))
        self._btn_cancel.clicked.connect(self._btn_cancel_clicked)

        self._btn_accept = QPushButton(self)
        self._btn_accept.setObjectName("_btn_accept")
        self._btn_accept.setGeometry(QRect(450, 440, 81, 24))
        self._btn_accept.clicked.connect(self._btn_accept_clicked)

        self._btn_box_path = QPushButton(self)
        self._btn_box_path.setObjectName("_btn_box_path")
        self._btn_box_path.setGeometry(QRect(550, 50, 75, 24))
        self._btn_box_path.clicked.connect(self._btn_box_path_clicked)

        self._btn_input_remove = QPushButton(self)
        self._btn_input_remove.setObjectName("_btn_input_remove")
        self._btn_input_remove.setGeometry(QRect(200, 190, 51, 24))
        self._btn_input_remove.clicked.connect(self._btn_input_remove_clicked)

        self._btn_input_add = QPushButton(self)
        self._btn_input_add.setObjectName("_btn_input_add")
        self._btn_input_add.setGeometry(QRect(260, 190, 51, 24))
        self._btn_input_add.clicked.connect(self._btn_input_add_clicked)

        self._btn_output_remove = QPushButton(self)
        self._btn_output_remove.setObjectName("_btn_output_remove")
        self._btn_output_remove.setGeometry(QRect(510, 190, 51, 24))
        self._btn_output_remove.clicked.connect(self._btn_output_remove_clicked)

        self._btn_output_add = QPushButton(self)
        self._btn_output_add.setObjectName("_btn_output_add")
        self._btn_output_add.setGeometry(QRect(570, 190, 51, 24))
        self._btn_output_add.clicked.connect(self._btn_output_add_clicked)

        self._fld_box_name = QLineEdit(self)
        self._fld_box_name.setObjectName("_fld_box_name")
        self._fld_box_name.setGeometry(QRect(130, 20, 411, 22))
        self._fld_box_name.textEdited.connect(self._test_data_accept_button)

        self._fld_box_path = QLineEdit(self)
        self._fld_box_path.setObjectName("_fld_box_path")
        self._fld_box_path.setGeometry(QRect(130, 50, 411, 22))
        self._fld_box_path.textEdited.connect(self._test_data_accept_button)

        self._fld_library = QLineEdit(self)
        self._fld_library.setObjectName("_fld_library")
        self._fld_library.setGeometry(QRect(130, 110, 411, 22))
        self._fld_library.textEdited.connect(self._test_data_accept_button)

        self._fld_creator = QLineEdit(self)
        self._fld_creator.setObjectName("_fld_creator")
        self._fld_creator.setGeometry(QRect(130, 140, 411, 22))
        self._fld_creator.textEdited.connect(self._test_data_accept_button)
        self._fld_creator.setText(self._user_data.user_name)

        self._cmb_box_type = QComboBox(self)
        self._cmb_box_type.setObjectName("_cmb_box_type")
        self._cmb_box_type.setGeometry(QRect(130, 80, 411, 22))
        self._cmb_box_type.currentIndexChanged.connect(self._test_data_accept_button)
        self._cmb_box_type.addItem('Composite', 0)
        self._cmb_box_type.addItem('Code', 1)

        self._tbl_inputs = QTableWidget(self)
        self._tbl_inputs.setObjectName("_tbl_inputs")
        self._tbl_inputs.setGeometry(QRect(10, 219, 300, 211))
        self._tbl_inputs.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self._tbl_inputs.selectionModel().selectionChanged.connect(self._tbl_inputs_selection_changed)
        self._tbl_inputs.setColumnCount(4)
        self._tbl_inputs.setHorizontalHeaderLabels(['Nom', 'Type', 'Nombre', 'Valeur'])
        self._tbl_inputs.currentCellChanged.connect(self._tbl_inputs_current_cell_changed)

        self._tbl_outputs = QTableWidget(self)
        self._tbl_outputs.setObjectName("_tbl_outputs")
        self._tbl_outputs.setGeometry(QRect(320, 220, 300, 211))
        self._tbl_outputs.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self._tbl_outputs.selectionModel().selectionChanged.connect(self._tbl_outputs_selection_changed)
        self._tbl_outputs.setColumnCount(4)
        self._tbl_outputs.setHorizontalHeaderLabels(['Nom', 'Type', 'Nombre', 'Valeur'])
        self._tbl_outputs.currentCellChanged.connect(self._tbl_outputs_current_cell_changed)

        self._lbl_box_name = QLabel(self)
        self._lbl_box_name.setObjectName("_lbl_box_name")
        self._lbl_box_name.setGeometry(QRect(10, 25, 111, 16))

        self._lbl_box_path = QLabel(self)
        self._lbl_box_path.setObjectName("_lbl_box_path")
        self._lbl_box_path.setGeometry(QRect(10, 55, 111, 16))

        self._lbl_box_type = QLabel(self)
        self._lbl_box_type.setObjectName("_lbl_box_type")
        self._lbl_box_type.setGeometry(QRect(10, 85, 111, 16))

        self._lbl_library = QLabel(self)
        self._lbl_library.setObjectName("_lbl_library")
        self._lbl_library.setGeometry(QRect(10, 115, 111, 16))

        self._lbl_creator = QLabel(self)
        self._lbl_creator.setObjectName("_lbl_creator")
        self._lbl_creator.setGeometry(QRect(10, 145, 111, 16))

        self._lbl_inputs = QLabel(self)
        self._lbl_inputs.setObjectName("_lbl_inputs")
        self._lbl_inputs.setGeometry(QRect(10, 195, 49, 16))

        self._lbl_outputs = QLabel(self)
        self._lbl_outputs.setObjectName("_lbl_outputs")
        self._lbl_outputs.setGeometry(QRect(320, 195, 49, 16))

        self._ln_separator = QFrame(self)
        self._ln_separator.setObjectName("_ln_separator")
        self._ln_separator.setGeometry(QRect(10, 170, 611, 21))
        self._ln_separator.setFrameShape(QFrame.HLine)
        self._ln_separator.setFrameShadow(QFrame.Sunken)

        QWidget.setTabOrder(self._fld_box_name, self._fld_box_path)
        QWidget.setTabOrder(self._fld_box_path, self._btn_box_path)
        QWidget.setTabOrder(self._btn_box_path, self._cmb_box_type)
        QWidget.setTabOrder(self._cmb_box_type, self._fld_library)
        QWidget.setTabOrder(self._fld_library, self._fld_creator)
        QWidget.setTabOrder(self._fld_creator, self._btn_accept)
        QWidget.setTabOrder(self._btn_accept, self._btn_cancel)
        QWidget.setTabOrder(self._btn_cancel, self._btn_input_remove)
        QWidget.setTabOrder(self._btn_input_remove, self._btn_input_add)
        QWidget.setTabOrder(self._btn_input_add, self._btn_output_remove)
        QWidget.setTabOrder(self._btn_output_remove, self._btn_output_add)
        QWidget.setTabOrder(self._btn_output_add, self._tbl_inputs)
        QWidget.setTabOrder(self._tbl_inputs, self._tbl_outputs)

        self.retranslateUi(self)

        QMetaObject.connectSlotsByName(self)

    def retranslateUi(self, Dialog) -> None :
        self.setWindowTitle(QCoreApplication.translate("_new_box_dialog", "Dialog", None))
        self._btn_cancel.setText(QCoreApplication.translate("_new_box_dialog", "Annuler", None))
        self._btn_accept.setText(QCoreApplication.translate("_new_box_dialog", "Accepter", None))
        self._btn_box_path.setText(QCoreApplication.translate("_new_box_dialog", "Rechercher", None))
        self._btn_input_remove.setText(QCoreApplication.translate("_new_box_dialog", "-", None))
        self._btn_input_add.setText(QCoreApplication.translate("_new_box_dialog", "+", None))
        self._btn_output_remove.setText(QCoreApplication.translate("_new_box_dialog", "-", None))
        self._btn_output_add.setText(QCoreApplication.translate("_new_box_dialogg", "+", None))
        self._lbl_inputs.setText(QCoreApplication.translate("n_new_box_dialog", "Entrées :", None))
        self._lbl_outputs.setText(QCoreApplication.translate("_new_box_dialog", "Sorties :", None))
        self._lbl_box_name.setText(QCoreApplication.translate("_new_box_dialog", "Nom de la box :", None))
        self._lbl_box_path.setText(QCoreApplication.translate("_new_box_dialog", "Chemin vers la box :", None))
        self._lbl_box_type.setText(QCoreApplication.translate("_new_box_dialog", "Type de box :", None))
        self._lbl_library.setText(QCoreApplication.translate("_new_box_dialog", "Bibliothèque :", None))
        self._lbl_creator.setText(QCoreApplication.translate("_new_box_dialog", "Créateur :", None))

    @pyqtSlot()
    def _test_data_accept_button (self) -> None :
        """Est la méthode pour activer/désactiver le bouton accept en fonction des champs d'édition."""
        
        if self._fld_box_name.text().strip().__len__() == 0 :
            self._btn_accept.setEnabled(False)
            return
        
        if self._fld_box_path.text().strip().__len__() == 0 :
            self._btn_accept.setEnabled(False)
            return

        if not os.path.isdir(self._fld_box_path.text()) :
            self._btn_accept.setEnabled(False)
            return

        if self._fld_library.text().strip().__len__() == 0 :
            self._btn_accept.setEnabled(False)
            return

        if self._fld_creator.text().strip().__len__() == 0 :
            self._btn_accept.setEnabled(False)
            return

        if self._tbl_inputs.rowCount() == 0 and self._tbl_outputs.rowCount() == 0 :
            self._btn_accept.setEnabled(False)
            return

        self._btn_accept.setEnabled(True)

    @pyqtSlot()
    def _btn_cancel_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton annuler."""
        self.close()

    @pyqtSlot()
    def _btn_box_path_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton pour rechercher un chemin."""
        path = QFileDialog.getExistingDirectory(self, "Recherche d'un dossier", os.getcwd())

        if path is None or path == '':
            return

        if not path.endswith('/') :
            path = path + '/'

        self._fld_box_path.setText(path)

    @pyqtSlot()
    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        super().keyPressEvent(event)

        if event.modifiers() == QtCore.Qt.ControlModifier:

            if event.key() == QtCore.Qt.Key_Z :
                self._actions_holder.undo()
                self._test_data_accept_button()

            elif event.key() == QtCore.Qt.Key_Y :
                self._actions_holder.redo()
                self._test_data_accept_button()

        elif event.modifiers() == QtCore.Qt.NoModifier :

            if event.key() == QtCore.Qt.Key_Delete :

                if self._tbl_inputs.hasFocus():
                    self._btn_input_remove_clicked()

                elif self._tbl_outputs.hasFocus():
                    self._btn_output_remove_clicked()

    @pyqtSlot()
    def _btn_input_add_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton pour ajouter une entrée."""

        addAction = NewBoxDialogAddAction(self._btn_input_add, 'Entrée', self._tbl_inputs)
        addAction.do()
        self._actions_holder.append(addAction)

        self._test_data_accept_button()
        self._btn_input_remove.setEnabled(self._tbl_inputs.selectedIndexes().__len__() > 0)

    @pyqtSlot()
    def _btn_input_remove_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton pour supprimer une entrée."""

        if self._tbl_inputs.selectedIndexes().__len__() == 0 :
            return

        removeAction = NewBoxDialogRemoveAction(self._btn_input_remove, self._tbl_inputs)
        removeAction.do()
        self._actions_holder.append(removeAction)
        self._btn_input_remove.setEnabled(self._tbl_inputs.selectedIndexes().__len__() > 0)

        self._test_data_accept_button()

    @pyqtSlot()
    def _btn_output_add_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton pour ajouter une sortie."""

        addAction = NewBoxDialogAddAction(self._btn_output_add, 'Sortie', self._tbl_outputs)
        addAction.do()
        self._actions_holder.append(addAction)

        self._test_data_accept_button()
        self._btn_output_remove.setEnabled(self._tbl_outputs.selectedIndexes().__len__() > 0)

    @pyqtSlot()
    def _btn_output_remove_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton pour supprimer une sortie."""

        if self._tbl_outputs.selectedIndexes().__len__() == 0 :
            return

        removeAction = NewBoxDialogRemoveAction(self._btn_output_remove, self._tbl_outputs)
        removeAction.do()
        self._actions_holder.append(removeAction)

        self._test_data_accept_button()
        self._btn_output_remove.setEnabled(self._tbl_outputs.selectedIndexes().__len__() > 0)
        
    @pyqtSlot()
    def _tbl_inputs_selection_changed (self) -> None :
        """Est la méthode appelée lorsque la sélection du tableau des entrée est modifiée."""

        self._btn_input_remove.setEnabled(self._tbl_inputs.selectedIndexes().__len__() > 0)

    @pyqtSlot()
    def _tbl_outputs_selection_changed (self) -> None :
        """Est la méthode appelée lorsque la sélection du tableau des entrée est modifiée."""

        self._btn_output_remove.setEnabled(self._tbl_outputs.selectedIndexes().__len__() > 0)
        
    @pyqtSlot()
    def _tbl_inputs_current_cell_changed (self) -> None :
        """Est la méthode appelée lorsqu'une valeur est modifiée dans le tableau de entrées."""
        LogManager().debug('_tbl_inputs_current_cell_changed')

    @pyqtSlot()
    def _tbl_outputs_current_cell_changed (self) -> None :
        """Est la méthode appelée lorsqu'une valeur est modifiée dans le tableau de sorties."""
        LogManager().debug('_tbl_outputs_current_cell_changed')

    @pyqtSlot()
    def _btn_accept_clicked (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur clique sur le bouton accepter."""

        boxModel : AbstractBoxModel = None

        lst_inputs : List[BoxInputModel] = []

        for row_index in range(self._tbl_inputs.rowCount()) :

            # ['Nom', 'Type', 'Nombre', 'Valeur']
            index : str =  self._tbl_inputs.item(row_index, 0).text()
            type : str =  self._tbl_inputs.cellWidget(row_index, 1).currentText()
            nbr : str =  self._tbl_inputs.cellWidget(row_index, 2).currentText()
            value : str =  self._tbl_inputs.item(row_index, 3).text()

            lst_inputs.append(BoxInputModel(index, get_data_type(type), QPointF(10, 10 + 40 * row_index), QSizeF(150, 30), 0))


        lst_outputs : List[BoxOutputModel] = []

        for row_index in range(self._tbl_outputs.rowCount()) :
            index : str =  self._tbl_outputs.item(row_index, 0).text()
            type : str =  self._tbl_outputs.cellWidget(row_index, 1).currentText()
            nbr : str =  self._tbl_outputs.cellWidget(row_index, 2).currentText()
            value : str =  self._tbl_outputs.item(row_index, 3).text()

            lst_inputs.append(BoxInputModel(index, get_data_type(type), QPointF(170, 10 + 40 * row_index), QSizeF(150, 30), 0))

        if self._cmb_box_type.currentData() == 0 :
            
            boxModel = BoxCompositeModel(
                self._fld_box_name.text(), 
                self._fld_box_path.text(), 
                self._fld_creator.text(),
                date.today(),
                0.1,
                lst_inputs,
                lst_outputs
            )

        else :

            boxModel = BoxCodeModel(
                self._fld_box_name.text(), 
                self._fld_box_path.text(), 
                self._fld_creator.text(),
                date.today(),
                0.1,
                '',
                lst_inputs,
                lst_outputs
            )

        self._value = boxModel
        self.close()


#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtCore import QPointF, QSizeF, Qt
from PyQt5.QtWidgets import QApplication, QGraphicsSceneHoverEvent
from muphyn.packages.interface.models.graphical_models.resizers.abstract_resizer import AbstractResizer

#-----------------------------------
# Class
#-----------------------------------

class HorizontalResizer (AbstractResizer) :
    """Est le resizer qui permet de modifier la longueur des boxes."""

    # -------------
    # Constructers
    # -------------

    def __init__ (self, parent) :
        
        AbstractResizer.__init__(self, parent, QSizeF(10, 20))

        self.setAcceptHoverEvents(True)

    # -------------
    # Methods
    # -------------

    def changeValue (self, value_: QPointF) -> QPointF :
        v : QPointF = value_ - self.pos()
        v.setY(0)
        return v

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        QApplication.setOverrideCursor(Qt.SizeHorCursor)
        return super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        while QApplication.overrideCursor() is not None:
            QApplication.restoreOverrideCursor()
        return super().hoverLeaveEvent(event)
        


#-----------------------------------
# Imports
#-----------------------------------

import os
import time
from typing import List, Any, Dict
from datetime import datetime

import matplotlib.pylab as plt

from PyQt5 import QtGui
from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtWidgets import QAction, QLabel, QMainWindow, QMenuBar, QSplitter, QToolBox, QWidget, QFileDialog, QApplication, QVBoxLayout, QGroupBox,QBoxLayout, QScrollArea
from PyQt5.QtGui import QKeySequence

from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.plci_core_scheduler_exception import SchedulerException
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.box_library.box_libraries_list import BoxLibrariesList
from muphyn.packages.interface.files.abstract_importer import AbstractImporter
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel
from muphyn.packages.interface.properties_pages.properties_widget import PropertiesWidget
from muphyn.packages.interface.dialogs.abstract_dialog import AbstractDialog
from muphyn.packages.interface.dialogs.dialogs_holder import DialogsHolder
from muphyn.packages.interface.editors.factory_editors import factory_editors
from muphyn.packages.interface.models.editable_models.abstract_diagram_model import AbstractDiagramModel
from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.project_tabs_holder import ProjectTabsHolder
from muphyn.packages.interface.recent_file_menu import RecentFileMenu
from muphyn.packages.interface.user_data.user_data import UserData
from muphyn.packages.interface.new_project_at_startup import new_project_on_startup
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.parser.parser import Parser
from muphyn.packages.interface.editors.diagram_editor import DiagramEditor
from muphyn.packages.interface.widgets.PopupWindows.YesNoPopup import YesNoMessageBox
from muphyn.packages.interface.utils.interface_constants import ApplicationWindowTitle
from muphyn.packages.interface.widgets.StandardActions import RotateLeftAction, RotateRightAction, UndoAction, RedoAction, DeleteAction, StartAction, StopAction

import muphyn.packages.interface.files.files as files

from muphyn.utils.paths import ROOT_DIR

#-----------------------------------
# Class
#-----------------------------------

class MainWindow (QMainWindow) :
    """Est la fenêtre principale."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        super(MainWindow, self).__init__()
        
        self.setGeometry(100, 100, 1024, 768)
        self.setWindowTitle(ApplicationWindowTitle)

        self._schedulers_libraries : SchedulersLibraries = SchedulersLibraries()
        
        # Load User parameters
        self._user_data : UserData = UserData("./user_data.yaml")
        self._user_data.load()

        # Load already added library
        for box_library in self._user_data.boxes_libraries :
            BoxesLibrariesManager().add_library(box_library)

        # Load schedulers libraries
        for scheduler_library in self._user_data.schedulers_libraries :
            self._schedulers_libraries.add_library(scheduler_library)

        # Load default libraries
        BoxesLibrariesManager().add_library(ROOT_DIR+"/libraries/box_library")
        self._schedulers_libraries.add_library(ROOT_DIR+"/libraries/scheduler_library")
        
        BoxesLibrariesManager().load_libraries()
        self._schedulers_libraries.load_libraries()

        self._dialogs_holder : DialogsHolder = DialogsHolder(self)
        self._dialogs_holder.dialog_opened_listener.append(self._dialog_opened_event)
        self._dialogs_holder.dialog_closed_listener.append(self._dialog_closed_event)
        self._dialogs_holder.dialog_closed_all_listener.append(self._dialogs_closed_all_event)

        self._opened_dialogs_actions : Dict[AbstractDialog, QAction] = {}

        self._init_actions()
        self._init_menu()
        self._init_ui()

        self.setAcceptDrops(True)

        self.regenerate_recent_files()
        self.open(new_project_on_startup(self._user_data, self._schedulers_libraries))

        # 
        if os.path.exists("backup.backup"):
            # Open Popup Window
            yesNoMessageBox = YesNoMessageBox("Nous avons détecté que l'application s'est fermé brutalement sans que vous ayez eu l'occasion de sauvegarder vos fichiers.\nVoulez-vous les récupérer ?", parent=self)
            yesNoMessageBox.show()

            yesNoMessageBox.accepted.connect(self.load_backup_files)
            yesNoMessageBox.rejected.connect(self.delete_backup_files)

    # -------------
    # Properties
    # -------------

    @property
    def dialogs_holder (self) -> DialogsHolder :
        """Permet de récuperer l'objet contenant les dialogues."""
        return self._dialogs_holder

    @property
    def schedulers_libraries (self) -> SchedulersLibraries :
        """Permet de récuperer l'objet contenant les bibliothèques de planificateurs."""
        return self._schedulers_libraries


    # -------------
    # Methods
    # -------------

    def dropEvent (self, drop_event : QtGui.QDropEvent) -> None :
        if drop_event.mimeData().hasUrls() :
            for url in drop_event.mimeData().urls() :
                self.open_file(url.toLocalFile())

        super().dropEvent(drop_event)

    def dragEnterEvent (self, drag_enter_event : QtGui.QDragEnterEvent) -> None :

        if drag_enter_event.mimeData().hasUrls() :
            drag_enter_event.accept()

        super().dragEnterEvent(drag_enter_event)

    def dragMoveEvent (self, drag_move_event : QtGui.QDragMoveEvent) -> None :
        return super().dragMoveEvent(drag_move_event)
    
    def open_file (self, path : str) -> None :
        """Permet d'ouvrir un fichier."""

        if os.path.exists(path) :
            editable_model : AbstractEditableModel = files.load(path)

            if editable_model is None :
                LogManager().error('no file to load !!')
                return

            else :
                self.open(editable_model)

    def open (self, adm : AbstractEditableModel) -> None :
        """Permet d'ouvrir un model d'édition."""

        # Check if a project with the same name already exists
        for graphical_editor_index in range(self._tabs.count()) :
            graphical_editor : AbstractEditableModel = self._tabs.widget(graphical_editor_index)

            if graphical_editor.editable_model.name == adm.name and graphical_editor.editable_model.path == adm.path :
                return
        
        # Create new tab
        tab : QWidget = factory_editors(self._tabs, adm)

        # Add Project to the LogManager
        LogManager().add_project_logger(tab.id, adm.name, adm.path)
        
        if isinstance(tab, QLabel) :
            self._tabs.addTab(tab, adm.name)

        else :
            self._tabs.addEditor(tab)

        if self._tabs.count() > 1 :
            self._tabs.setCurrentIndex(self._tabs.count() - 1)

        # Add opened file to recent files
        self.add_recent_file(adm.path + adm.name + '.yaml')

    def add_recent_file (self, path : str) -> None :
        """Permet d'essayer d'ajouter le fichier dans les fichiers récents."""

        if path in self._user_data.recent_files :
            self._user_data.recent_files.remove(path)
        
        self._user_data.recent_files.insert(0, path)
        self.regenerate_recent_files()

    def regenerate_recent_files (self) -> None : 
        """Permet de regénérer le menu des fichiers récents."""
        max_files = 10
        current_file_number = 0
        for action_index in range(len(self._mn_recent_files.actions())).__reversed__() :
            self._mn_recent_files.removeAction(self._mn_recent_files.actions()[action_index])

        for path in self._user_data.recent_files :

            if current_file_number >= max_files : 
                return

            if os.path.exists(path) :
                mn_recent_file_action = RecentFileMenu(self, path, self.open_file)
                self._mn_recent_files.addAction(mn_recent_file_action)

                current_file_number += 1

    @pyqtSlot()
    def _file_open_method (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur veut ouvrir un fichier."""

        path = QFileDialog.getOpenFileName(self, f'{ApplicationWindowTitle} - Open project file', os.getcwd(), 'YAML (*.yaml)')[0]
        if path is None or len(path) == 0 :
            return

        self.open_file(path)

    @pyqtSlot()
    def _file_save_method (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur veut sauvegarder le fichier en cours."""

        if self._tabs.current_editor is None : 
            return

        if self._tabs.current_editor.editable_model is None : 
            return
        
        file_name = self._tabs.current_editor.editable_model.path.__str__()
        if not(file_name.endswith(os.sep)) :
            file_name = file_name + os.sep

        file_name = file_name + self._tabs.current_editor.editable_model.name

        if not(file_name.endswith('.yaml')):
            file_name = file_name + '.yaml'

        files.save(self._tabs.current_editor.editable_model, file_name)

    @pyqtSlot()
    def _file_save_as_method (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur veut sauvegarder le fichier en cours sous un nouveau nom."""

        if self._tabs.current_editor is None : 
            return

        if self._tabs.current_editor.editable_model is None : 
            return

        path = QFileDialog.getSaveFileName(self, f'{ApplicationWindowTitle} - Save as file', os.getcwd(), 'YAML (*.yaml)')[0]

        if path is None or len(path) == 0 :
            return

        if files.save(self._tabs.current_editor.editable_model, path) : 

            file_name = os.path.basename(path)
            abs_path = path[:-file_name.__len__()]
            file_name = file_name[:-('.yaml').__len__()]

            self._tabs.current_editor.editable_model.set_path_name(abs_path, file_name)

    @pyqtSlot()
    def _file_close_method (self) -> None : 
        """Est la méthode appelée pour fermer le fichier actuel."""
        self.tab_close_request(self._tabs.currentIndex())

    @pyqtSlot()
    def _undo_method (self) -> None :
        """Est la méthode appelée pour défaire la dernière action."""

        if self._tabs.current_editor is None : 
            return

        self._tabs.current_editor.actions_holder.undo()

    @pyqtSlot()
    def _redo_method (self) -> None : 
        """Est la méthode appelée pour refaire la dernière action défaite."""

        if self._tabs.current_editor is None : 
            return

        self._tabs.current_editor.actions_holder.redo()

    @pyqtSlot()
    def _window_libraries_method (self) -> None :
        """Est la méthode appelée pour afficher un dialogue permettant à l'utilisateur de modifier les bibliothèques."""
        self.dialogs_holder.show_dialog('library', True, boxes_libraries = BoxesLibrariesManager(), schedulers_libraries = self._schedulers_libraries)

    @pyqtSlot()
    def _help_documentation_method (self) -> None :
        """Est la méthode appelée pour afficher un dialogue permettant d'aider l'utilisateur."""
        self.dialogs_holder.show_dialog('documentation', False)

    @pyqtSlot()
    def _help_about_method (self) -> None :
        """Est la méthode appelée pour afficher un dialogue affichant les détails des licences et des personnes ayant travaillés sur le programme."""
        self.dialogs_holder.show_dialog('about', False)

    @pyqtSlot()
    def _file_new_simulation_method (self) -> None :
        """Est la méthode appellée lorsque l'utilisateur veut créer une nouvelle simulation."""
        self.dialogs_holder.show_dialog('new_simulation', True, schedulers_libraries = self._schedulers_libraries)

    @pyqtSlot()
    def _simulation_parameters_method (self) -> None :
        """Est la méthode appelée pour afficher un digalogue permettant de modifier les paramètres de la simulation."""
        if isinstance(self._tabs.current_editor.editable_model, SimulationModel) : 
            self.dialogs_holder.show_dialog('simulation_parameters_dialog', True, current_simulation = self._tabs.current_editor.editable_model, schedulers_libraries = self._schedulers_libraries)

    @pyqtSlot()
    def _simulation_launch_method (self) -> None :
        """Est la méthode appelée pour lancer la simulation en cours d'édition."""

        global_simulation_start = time.perf_counter()
        
        if self._tabs.current_editor is None : 
            return

        if self._tabs.current_editor.editable_model is None : 
            return

        parser = Parser()
        scheduler : Scheduler = parser.parse(self._tabs.current_editor.editable_model, self.schedulers_libraries)
        
        if scheduler is None : 
            LogManager().error('No scheduler returned for the simulation !!')

        else :
            start_time = time.perf_counter()

            schedulerException : SchedulerException = scheduler.schedule()
            stop_time = time.perf_counter()

            if schedulerException is None :
                LogManager().debug(f"The simulation is correctly finished and it took : {stop_time - start_time}s.")
                LogManager().debug(f"The simulation + initialization took : {stop_time - global_simulation_start}s.")

                plt.show()
                if len(plt.get_fignums()) > 0 :
                    LogManager().debug('figures shown : ')
                    for i in plt.get_fignums() :
                        LogManager().debug(f' - {i}')

            else :
                LogManager().error(f"The simulation encoutered an error after {stop_time - start_time: .3f}s of simulation.")
                schedulerException.print()

    @pyqtSlot()
    def _file_new_box_method (self) -> None :
        """Est la méthode appellée lorsque l'utilisateur veut créer une nouvelle box."""
        self.dialogs_holder.show_dialog('new_box', True, user_data = self._user_data)

    @pyqtSlot()
    def _file_new_solver_method (self) -> None :
        """Est la méthode appellée lorsque l'utilisateur veut créer un nouveau solveur."""
        self.dialogs_holder.show_dialog('new_scheduler', True, user_data = self._user_data)

    @pyqtSlot()
    def _edit_copy_method (self) -> None :
        """Est la méthode appelée pour copier la sélection actuelle."""
        if self._tabs.current_editor is None : 
            return

        self._tabs.current_editor.copy()

    @pyqtSlot()
    def _edit_cut_method (self) -> None :
        """Est la méthode appelée pour couper la sélection actuelle."""
        if self._tabs.current_editor is None : 
            return

        self._tabs.current_editor.cut()

    @pyqtSlot()
    def _edit_paste_method (self) -> None :
        """Est la méthode appelée pour coller le contenu du clipboard."""
        if self._tabs.current_editor is None : 
            return

        self._tabs.current_editor.paste()

    @pyqtSlot()    
    def _edit_rot_clock_90_method (self) -> None :
        """Est la méthode appelée pour faire tourner la box sélectionné de 90° dans le sens des aiguilles d'une montre."""
        
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        for element in current_editor.selected_elements() :
            if hasattr(element, 'setRotation') :
                element.setRotation(element.rotation() + 90)

        
    @pyqtSlot()
    def _edit_rot_anti_clock_90_method (self) -> None :
        """Est la méthode appelée pour faire tourner la box sélectionné de 90° dans le sens des aiguilles d'une montre."""
        
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        for element in current_editor.selected_elements() :
            if hasattr(element, 'setRotation') :
                element.setRotation(element.rotation() - 90)

    @pyqtSlot()
    def _edit_rot_reverse_method (self) -> None :
        """Est la méthode appelée pour retourner la box sélectionné."""
        
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        for element in current_editor.selected_elements() :
            if hasattr(element, 'setRotation') :
                element.setRotation(element.rotation() + 180)

    @pyqtSlot()
    def _edit_select_all_method (self) -> None :
        """Est la méthode appelée pour sélectionner tous les éléments."""
        
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        for element in current_editor.elements() :
            element.setSelected(True)

    @pyqtSlot()
    def _edit_reverse_selection_method (self) -> None :
        """Est la méthode appelée pour inverser la sélection des éléments."""
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        for element in current_editor.elements() :
            element.setSelected(not(element.isSelected()))
    
    @pyqtSlot()
    def _edit_erase_selection_method (self) -> None :
        """Est la méthode appelée pour supprimer la sélection."""
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return
        
        current_editor.delete_selection()
    
    @pyqtSlot()
    def _window_zoom_method (self) -> None : 
        """Est la méthode appelée pour zoomer dans l'interface d'édition."""
        
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        current_editor.zoom(1)
    
    @pyqtSlot()
    def _window_dezoom_method (self) -> None :
        """Est la méthode appelée pour dézoomer dans l'interface d'édition."""
        
        current_editor = self._tabs.current_editor

        if current_editor is None :
            return

        current_editor.zoom(-1)

    def _init_actions (self) -> None :
        """Permet de créer les actions que la fenêtres peut réaliser."""

        self._file_new_simulation_action = QAction('&Simulation', self)
        self._file_new_simulation_action.triggered.connect(self._file_new_simulation_method)
        self._file_new_simulation_action.setShortcut(QKeySequence('Ctrl+N'))

        self._file_new_box_action = QAction('&Box', self)
        self._file_new_box_action.triggered.connect(self._file_new_box_method)
        self._file_new_box_action.setShortcut(QKeySequence('Ctrl+Shift+N'))

        self._file_new_solver_action = QAction('Sche&duler', self)
        self._file_new_solver_action.triggered.connect(self._file_new_solver_method)

        self._file_open_action = QAction('&Open', self)
        self._file_open_action.triggered.connect(self._file_open_method)
        self._file_open_action.setShortcut(QKeySequence('Ctrl+O'))
        
        self._file_save_action = QAction('&Save', self)
        self._file_save_action.triggered.connect(self._file_save_method)
        self._file_save_action.setShortcut(QKeySequence('Ctrl+S'))

        self._file_save_as_action = QAction('S&ave as', self)
        self._file_save_as_action.triggered.connect(self._file_save_as_method)
        self._file_save_as_action.setShortcut(QKeySequence('Ctrl+Shift+S'))

        self._file_export_action = QAction('&Export', self)
        self._file_export_action.triggered.connect(self._file_export_method)
        self._file_export_action.setShortcut(QKeySequence('Ctrl+E'))

        self._file_close_action = QAction('&Close', self)
        self._file_close_action.triggered.connect(self._file_close_method)
        self._file_close_action.setShortcut(QKeySequence('Ctrl+W'))

        self._file_exit_action = QAction('&Quit', self)
        self._file_exit_action.triggered.connect(self.close)
        self._file_exit_action.setShortcut(QKeySequence('Alt+F4'))

        self._edit_undo_action = UndoAction('&Undo', self)
        self._edit_undo_action.triggered.connect(self._undo_method)
        self._edit_undo_action.setShortcut(QKeySequence('Ctrl+Z'))

        self._edit_redo_action = RedoAction('&Redo', self)
        self._edit_redo_action.triggered.connect(self._redo_method)
        self._edit_redo_action.setShortcut(QKeySequence('Ctrl+Y'))

        self._edit_cut_action = QAction('C&ut', self)
        self._edit_cut_action.triggered.connect(self._edit_cut_method)
        self._edit_cut_action.setShortcut(QKeySequence('Ctrl+X'))

        self._edit_copy_action = QAction('&Copy', self)
        self._edit_copy_action.triggered.connect(self._edit_copy_method)
        self._edit_copy_action.setShortcut(QKeySequence('Ctrl+C'))

        self._edit_paste_action = QAction('&Paste', self)
        self._edit_paste_action.triggered.connect(self._edit_paste_method)
        self._edit_paste_action.setShortcut(QKeySequence('Ctrl+V'))

        self._rot_clock_90_action = RotateRightAction('Rotate right', self)
        self._rot_clock_90_action.triggered.connect(self._edit_rot_clock_90_method)

        self._rot_anti_clock_90_action = RotateLeftAction('Rotate left', self)
        self._rot_anti_clock_90_action.triggered.connect(self._edit_rot_anti_clock_90_method)

        self._rot_reverse_action = QAction('Flip', self)
        self._rot_reverse_action.triggered.connect(self._edit_rot_reverse_method)

        self._select_all_action = QAction('Select all', self)
        self._select_all_action.triggered.connect(self._edit_select_all_method)
        self._select_all_action.setShortcut(QKeySequence('Ctrl+A'))

        self._reverse_selection_action = QAction('Invert selection', self)
        self._reverse_selection_action.triggered.connect(self._edit_reverse_selection_method)
        self._reverse_selection_action.setShortcut(QKeySequence('Ctrl+Shift+A'))

        self._edit_new_simulation_from_selection_action = QAction('Create simulation from selection', self)
        self._edit_new_simulation_from_selection_action.triggered.connect(self._edit_new_simulation_from_selection_method)

        self._edit_new_composite_box_from_selection_action = QAction('Create composite box from selection', self)
        self._edit_new_composite_box_from_selection_action.triggered.connect(self._edit_new_composite_box_from_selection_method)

        self._edit_erase_selection_action = DeleteAction('Delete', self)
        self._edit_erase_selection_action.triggered.connect(self._edit_erase_selection_method)
        self._edit_erase_selection_action.setShortcut(QKeySequence('Delete'))

        self._window_zoom_action = QAction('Zoom +', self)
        self._window_zoom_action.triggered.connect(self._window_zoom_method)
        self._window_zoom_action.setShortcut(QKeySequence('Ctrl++'))

        self._window_dezoom_action = QAction('Zoom -', self)
        self._window_dezoom_action.triggered.connect(self._window_dezoom_method)
        self._window_dezoom_action.setShortcut(QKeySequence('Ctrl+-'))

        self._window_close_all_action = QAction('Close dialog windows', self)
        self._window_close_all_action.triggered.connect(self._dialogs_holder.close_all)
        self._window_close_all_action.setEnabled(self._dialogs_holder.len > 0)

        self._window_libraries_action = QAction('Libraries', self)
        self._window_libraries_action.triggered.connect(self._window_libraries_method)
        
        self._window_options_action = QAction('Options', self)
        self._window_options_action.triggered.connect(self._window_options_method)

        self._simulation_parameters_action = QAction('Parameters', self)
        self._simulation_parameters_action.triggered.connect(self._simulation_parameters_method)

        self._simulation_launch_action = StartAction('Start simulation', self)
        self._simulation_launch_action.triggered.connect(self._simulation_launch_method)
        self._simulation_launch_action.setShortcut(QKeySequence('F5'))

        self._simulation_stop_action = StopAction('Stop simulation', self)
        self._simulation_stop_action.triggered.connect(self._simulation_stop_method)
        self._simulation_stop_action.setShortcut(QKeySequence('Shift+F5'))

        self._help_documentation_action = QAction('&Documentation', self)
        self._help_documentation_action.triggered.connect(self._help_documentation_method)
        self._help_documentation_action.setShortcut(QKeySequence('F1'))
        
        self._help_about_action = QAction('&About', self)
        self._help_about_action.triggered.connect(self._help_about_method)

    def _init_menu (self) -> None : 
        """Permet de créer le menu afficher dans la fenêtre."""

        self._menu_bar = QMenuBar()
        self.setMenuBar(self._menu_bar)

        self._mn_file = self._menu_bar.addMenu('&Files')
        self._mn_file_new = self._mn_file.addMenu('&New')
        self._mn_file_new.addAction(self._file_new_simulation_action)
        self._mn_file_new.addAction(self._file_new_box_action)
        self._mn_file_new.addAction(self._file_new_solver_action)
        self._mn_file.addAction(self._file_open_action)
        self._mn_file.addAction(self._file_save_action)
        self._mn_file.addAction(self._file_save_as_action)
        self._mn_recent_files = self._mn_file.addMenu("Recent Files")
        self._mn_file.addSeparator()
        self._mn_file.addAction(self._file_export_action)
        self._mn_file.addSeparator()
        self._mn_file.addAction(self._file_close_action)
        self._mn_file.addSeparator()
        self._mn_file.addAction(self._file_exit_action)

        self._mn_edit = self._menu_bar.addMenu('&Edit')
        self._mn_edit.addAction(self._edit_undo_action)
        self._mn_edit.addAction(self._edit_redo_action)
        self._mn_edit.addSeparator()
        self._mn_edit.addAction(self._edit_cut_action)
        self._mn_edit.addAction(self._edit_copy_action)
        self._mn_edit.addAction(self._edit_paste_action)
        self._mn_edit.addSeparator()
        self._mn_edit_rotation = self._mn_edit.addMenu('Rotate')
        self._mn_edit_rotation.addAction(self._rot_clock_90_action)
        self._mn_edit_rotation.addAction(self._rot_anti_clock_90_action)
        self._mn_edit_rotation.addAction(self._rot_reverse_action)
        self._mn_edit.addSeparator()
        self._mn_edit.addAction(self._select_all_action)
        self._mn_edit.addAction(self._reverse_selection_action)
        self._mn_edit.addSeparator()
        self._mn_edit.addAction(self._edit_new_simulation_from_selection_action)
        self._mn_edit.addAction(self._edit_new_composite_box_from_selection_action)
        self._mn_edit.addSeparator()
        self._mn_edit.addAction(self._edit_erase_selection_action)

        self._mn_window = self._menu_bar.addMenu('&View')
        self._mn_window.addAction(self._window_zoom_action)
        self._mn_window.addAction(self._window_dezoom_action)
        self._mn_window.addSeparator()
        self._mn_window_dialogues = self._mn_window.addMenu('Opened dialog windows')
        self._mn_window_dialogues.setEnabled(self._dialogs_holder.len > 0)
        self._mn_window.addAction(self._window_close_all_action)
        self._mn_window.addSeparator()
        self._mn_window.addAction(self._window_libraries_action)
        self._mn_window.addAction(self._window_options_action)
        
        self._mn_simulation = self._menu_bar.addMenu('&Simulation')
        self._mn_simulation.addAction(self._simulation_parameters_action)
        self._mn_simulation.addSeparator()
        self._mn_simulation.addAction(self._simulation_launch_action)
        self._mn_simulation.addAction(self._simulation_stop_action)

        self._mn_help = self._menu_bar.addMenu('&?')
        self._mn_help.addAction(self._help_documentation_action)
        self._mn_help.addSeparator()
        self._mn_help.addAction(self._help_about_action)

    def tabs_elements_selected_changed (self, elements_i) -> None :
        """Est la méthode appelée lorsque l'utilisateur modifie sa sélection."""
        # Get Current editor instance
        current_editor = self._tabs.current_editor

        # If no current editor -> error
        if current_editor is None :
            LogManager().error("MainWindow.tabs_elements_selected_changed: Trying to select current editor but get None")
            return
        
        # Get all selected elements
        elements_l: list = list(elements_i())

        # If only one element is selected
        #   Note : If more than one element selected we can't get the last selected one
        #           because the elements are in the list in the same order of creation
        if len(elements_l) == 1 :
            self._tool_box_page_properties.actions_holder = current_editor.actions_holder
            self._tool_box_page_properties.current_model = elements_l[0]

        else :
            self._tool_box_page_properties.actions_holder = None
            self._tool_box_page_properties.current_model = None
    
    def _init_ui (self) :
        """Est la méthode appelée pour afficher les éléments visuels."""

        # Main Window Splitter
        self._splt_main_window = QSplitter(Qt.Orientation.Horizontal)
        self._splt_main_window.setChildrenCollapsible(False)
        self._splt_main_window.setContentsMargins(5, 5, 5, 5)

        # Vertical Splitter
        self._splitter_lib_prop = QSplitter(Qt.Orientation.Vertical)
        self._splitter_lib_prop.setMinimumWidth(355)
        self._splitter_lib_prop.setChildrenCollapsible(False)

        # Project tab holder
        self._tabs = ProjectTabsHolder(self._splt_main_window)
        self._tabs.elements_selected_changed.connect(self.tabs_elements_selected_changed)
        self._tabs.tabCloseRequested.connect(self.tab_close_request)

        # Init Libraries list widget
        lib_groupbox = QGroupBox("Libraries")
        self._libraries = BoxLibrariesList()
        lib_groupbox.setLayout(self._libraries.layout())

        # Init Properties list widget
        properties_groupbox = QGroupBox("Properties")
        self._tool_box_page_properties = PropertiesWidget()
        boxLayout = QVBoxLayout()

        # Init Scroll Area
        scrollArea = QScrollArea()
        scrollArea.setWidgetResizable(True)
        scrollArea.setBaseSize
        scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        # Add Properties Widget to Scroll Area
        scrollArea.setWidget(self._tool_box_page_properties)

        # Add Scroll Area to Box Layout
        boxLayout.addWidget(scrollArea)

        # Add Properties Box Layout to GroupBox
        properties_groupbox.setLayout(boxLayout.layout())

        # Add Items in right Vertical Splitter
        self._splitter_lib_prop.addWidget(lib_groupbox)
        self._splitter_lib_prop.addWidget(properties_groupbox)
        self._splitter_lib_prop.setStretchFactor(0, 0)
        self._splitter_lib_prop.setStretchFactor(1, 1)

        # Add Item in Main Window Splitter
        self._splt_main_window.addWidget(self._tabs)
        self._splt_main_window.addWidget(self._splitter_lib_prop)
        self._splt_main_window.setStretchFactor(0, 1)
        self._splt_main_window.setStretchFactor(1, 0)

        # Set Splitter as main widget
        self.setCentralWidget(self._splt_main_window)

    @pyqtSlot(int)
    def tab_close_request (self, index : int = ...) :
        
        if self._tabs.count() == 1 :
            return

        # Get current widget
        current_widget = self._tabs.widget(index)

        # Remove project in LogManager
        LogManager().remove_project_logger(current_widget.id)

        # Remove tab from Tabs Holder
        self._tabs.removeTab(index)

        # Delete Widget
        current_widget.deleteLater()



    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """Est la méthode appelée à la fermeture de la fenêtre."""
        super().closeEvent(event)

        self._user_data.save()

    def _dialog_opened_event (self, abstract_dialog : AbstractDialog) -> None :
        """Est la méthode appelée lorsqu'un dialogue est ouvert."""

        self._mn_window_dialogues.setEnabled(self._dialogs_holder.len > 0)
        self._window_close_all_action.setEnabled(self._dialogs_holder.len > 0)

        new_dialog_action : QAction = QAction(abstract_dialog.windowTitle(), self)
        new_dialog_action.triggered.connect(lambda : abstract_dialog.close())
        self._opened_dialogs_actions[abstract_dialog] = new_dialog_action
        self._mn_window_dialogues.addAction(new_dialog_action)

    def _dialog_closed_event (self, abstract_dialog : AbstractDialog, arg : Any) -> None :
        """Est la méthode appelée lorsqu'un dialogue est fermé.""" 

        self._mn_window_dialogues.setEnabled(self._dialogs_holder.len > 0)
        self._window_close_all_action.setEnabled(self._dialogs_holder.len > 0)

        self._mn_window_dialogues.removeAction(self._opened_dialogs_actions[abstract_dialog])
        del self._opened_dialogs_actions[abstract_dialog]

        if abstract_dialog.name == 'library':
            self._user_data.schedulers_libraries.clear()
            self._user_data.boxes_libraries.clear()

            for library in BoxesLibrariesManager()._libraries:
                if not library.path in self._user_data.boxes_libraries :
                    self._user_data.boxes_libraries.append(library.path)

            self._libraries.libraries_reloaded()

            for library in self._schedulers_libraries._libraries:
                if not library.path in self._user_data.schedulers_libraries :
                    self._user_data.schedulers_libraries.append(library.path)

        elif abstract_dialog.name == 'new_simulation' :

            if abstract_dialog.value is None :
                return

            self.open(abstract_dialog.value)

        elif abstract_dialog.name == 'new_box' :

            if abstract_dialog.value is None :
                return

            self.open(abstract_dialog.value)

        elif abstract_dialog.name == 'simulation_parameters_dialog' :

            if abstract_dialog.value is None : 
                return

            if self._tabs.current_editor is None : 
                return

            if not(isinstance(self._tabs.current_editor.editable_model, SimulationModel)) :
                return

            if isinstance(abstract_dialog.value, SchedulerModel) :
                self._tabs.current_editor.editable_model.scheduler_model = abstract_dialog.value

        else :
            LogManager().error(f"dialog closed not supported : {abstract_dialog.name}")


    def _dialogs_closed_all_event (self, dialogs_holder : DialogsHolder) -> None :
        """Est la méthode appelée lorsque toutes les boite de dialogues ont été fermée."""
        
        self._mn_window_dialogues.setEnabled(self._dialogs_holder.len > 0)
        self._window_close_all_action.setEnabled(self._dialogs_holder.len > 0)

        for dialog in self._opened_dialogs_actions :
            self._mn_window_dialogues.removeAction(self._opened_dialogs_actions[dialog])

        self._opened_dialogs_actions.clear()

    def backup_save_projects(self):
        # Init Bakcup Files list 
        backup_files = []
        for tab in self._tabs:
            # Get Project Directory
            dirname = tab.editable_model.path.__str__()

            # Build Project backup file name
            filename = f"{tab.editable_model.name}_backup.yaml"

            # Save backup file
            bakcup_file_path = os.path.join(dirname, filename)
            files.save(tab.editable_model, bakcup_file_path)

            # Append backup file path to list
            backup_files.append(bakcup_file_path)
        
        # Save Bakcup Files list in main application directory
        with open("backup.backup", "w") as f:
            f.write("\n".join(backup_files))

    def load_backup_files(self):
        # Load Backup files
        with open("backup.backup", "r") as f:
            backup_files = [backup_file_path for backup_file_path in f.readlines()]
            for backup_file in backup_files:
                self.open_file(backup_file)
                
        # Delete Backup files
        self.delete_backup_files()

    def delete_backup_files(self):
        os.remove("backup.backup")

    @pyqtSlot()
    def _file_export_method (self) -> None :
        """Est la méthode appelée pour exporter le fichier actuel."""
        LogManager().error('Not implemented : Export')
    
    @pyqtSlot()
    def _edit_new_simulation_from_selection_method (self) -> None :
        """Est la méthode appelée pour créer une nouvelle simulation depuis la sélection."""
        LogManager().error('Not implemented : New simulation from selection')

    @pyqtSlot()
    def _edit_new_composite_box_from_selection_method (self) -> None :
        """Est la méthode appelée pour créer une nouvelle box composite depuis la sélection."""
        LogManager().error('Not implemented : New composite box from selection')
        
    @pyqtSlot()
    def _window_options_method (self) -> None :
        """Est la méthode appelée pour afficher un dialogue permettant à l'utilisateur de gérer les options du programme."""
        LogManager().error('Not implemented : Show options popup')

    @pyqtSlot()
    def _simulation_stop_method (self) -> None :
        """Est la méthode appelée pour stopper l'éventuel simulation en cours de simulation."""
        LogManager().error('Not implemented : Stop simulate')



import sys
from PyQt5.QtWidgets import QApplication, QWidget, QComboBox
from PyQt5.QtCore import QRect, QPointF, QSizeF
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibraries
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.editors.diagram_editor import construct_box
from muphyn.packages.interface.models.graphical_models.box_input_model import BoxInputModel
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.graphical_models.box_output_model import BoxOutputModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel

from muphyn.packages.interface.properties_pages.properties_widget import PropertiesWidget

from .unit_test import get_boxes_libraries

boxes_libraries : BoxesLibraries = get_boxes_libraries()

def get_box (library, name) -> BoxModel :

    box_data : AbstractBoxData = boxes_libraries.get_box_data(library, name)

    if box_data is None : 
        return None 
    
    return construct_box(box_data, QPointF(0, 0), True)
     

false_input_box = BoxModel('', 'false_box', QPointF(0, 0), QSizeF(0, 0), 0, False, '')
false_input = InputConnectionModel('false_input', DataType.FLOAT, QPointF(0, 0))
false_input_box._inputs.append(false_input)

false_output_box = BoxModel('', '', QPointF(0, 0), QSizeF(0, 0), 0, False, '')
false_output = OutputConnectionModel('false_output', DataType.FLOAT, QPointF(0, 0))
false_output_box._outputs.append(false_output)
link_element = false_output.add_link(false_input)

elements = [
    link_element, 
    BoxInputModel('', DataType.FLOAT, QPointF(0, 0), QSizeF(0, 0)), 
    BoxOutputModel('', DataType.FLOAT, QPointF(0, 0), QSizeF(0, 0)),
    get_box('Boxes.Math', 'Addition'),
    get_box('Boxes.Math', 'Amplifier'),
    get_box('Boxes.Sources', 'Constant'),
    get_box('Boxes.Math', 'Derivator'),
    get_box('Boxes.Sinks', 'Graph'),
    get_box('Boxes.Math', 'Integrator'),
    get_box('Boxes.Math', 'Multiplier'),
    get_box('Boxes.Sources', 'Ramp'),
    get_box('Boxes.Sources', 'Sine'),
    get_box('Boxes.Sources', 'Square'),
    get_box('Boxes.Sources', 'Step')
]

class test_windows (QWidget) :

    def __init__ (self) :

        QWidget.__init__(self, None)
        self.init_ui()
        self.resize(330, 400)

    def init_ui (self) :
        self.cmb_elements = QComboBox(self)
        self.cmb_elements.setObjectName(u"comboBox")
        self.cmb_elements.setGeometry(QRect(10, 10, 311, 22))

        self.cmb_elements.addItem('Link')
        self.cmb_elements.addItem('Composite box input')
        self.cmb_elements.addItem('Composite box output')
        self.cmb_elements.addItem('Box - Addition')
        self.cmb_elements.addItem('Box - Amplifier')
        self.cmb_elements.addItem('Box - Constant')
        self.cmb_elements.addItem('Box - Derivator')
        self.cmb_elements.addItem('Box - Graph')
        self.cmb_elements.addItem('Box - Integrator')
        self.cmb_elements.addItem('Box - Multiplier')
        self.cmb_elements.addItem('Box - Ramp')
        self.cmb_elements.addItem('Box - Sine')
        self.cmb_elements.addItem('Box - Square')
        self.cmb_elements.addItem('Box - Step')

        self.properties_widget = PropertiesWidget(self)
        self.properties_widget.setObjectName(u"widget")
        self.properties_widget.setGeometry(QRect(10, 40, 311, 351))
        
        self.cmb_elements.setCurrentIndex(2)
        self.cmb_elements.currentIndexChanged.connect(self.cmb_elements_current_index_changed)
        self.cmb_elements.setCurrentIndex(3)

    def cmb_elements_current_index_changed (self) -> None : 
        index = self.cmb_elements.currentIndex()
        element = elements[index]
        
        self.properties_widget.current_model = element


app = QApplication(sys.argv)
win = test_windows()
win.show()

sys.exit(app.exec_())


#-----------------------------------
# Imports
#-----------------------------------
 
import os
import pathlib
from datetime import date

from PyQt5.QtCore import QCoreApplication

from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.user_data.user_data import UserData
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.interface.models.editable_models.scheduler_model import SchedulerModel

#-----------------------------------
# Functions
#-----------------------------------

def new_project_on_startup (user_data : UserData, schedulers_libraries : SchedulersLibraries) -> SimulationModel :
    """Permet de créer un nouveau projet vide au démarrage du projet."""

    name = QCoreApplication.translate("_new_project_on_startup", "New Project", None)
    path = pathlib.Path(__file__).parent.parent.parent.absolute().__str__() + os.sep.__str__()
    
    creator = user_data.user_name
    _date = date.today() 
    version = 1.0

    step_time = 0.1
    stop_time = 10
    
    scheduler = None
    for sch in schedulers_libraries.schedulers :
        scheduler = sch
        break

    scheduler_params = SchedulerParams(stop_time, step_time)
    if not(scheduler is None) :
        scheduler_model : SchedulerModel = SchedulerModel(scheduler.scheduler_library, scheduler.scheduler_name, scheduler_params)
    else : 
        scheduler_model : SchedulerModel = SchedulerModel('', '', scheduler_params)
    
    return SimulationModel(name, path, creator, _date, version, scheduler_model)


#-----------------------------------
# Imports
#-----------------------------------
from PyQt5.QtCore import QCoreApplication, Qt
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import QLabel

from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.properties_pages.data_property_widgets import property_widget_factory
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel


#-----------------------------------
# Class
#-----------------------------------

class ParameterPropertiesEditor (AbstractPropertiesEditor) :
    """Est l'éditeur qui permet modifier un paramètre."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_model : BoxModel, parameter : str) :

        self._parameter = parameter
        self._box_model = box_model
        self._box_model.param_changed.connect(self.box_param_changed)

        AbstractPropertiesEditor.__init__(self, box_model)

        parameter_to_edit = self.box_model.get_parameter(self._parameter)
        self._semaphore_values = True

        self._widget_value_editor.setValue(parameter_to_edit['value'])

        self._semaphore_values = False
    # -------------
    # Properties
    # -------------

    @property
    def parameter (self) -> str :
        """Permet de récuperer le nom du paramètre qui doit être édité dans la box."""
        return self._parameter

    @property
    def box_model (self) -> BoxModel : 
        """Permet de récuperer le modème de la box dont le paramètre est édité."""
        return self._box_model

    # -------------
    # Methods
    # -------------
    def box_param_changed (self, box, param_name, old_value, new_value) -> None :
        """Est la méthode qui est appelée lorsque la box modifiée par le panel voit un de ses paramètres changer."""

        LogManager().debug('properties editor . box params changed')
        LogManager().debug(f'- param name : {param_name}')
        LogManager().debug(f'- param value : {new_value}')
        LogManager().debug(f'- param last value : {old_value}')

        if not(box == self._model) :
            LogManager().error('box is not the same')
            return

        if not(param_name == 'params.' + self._parameter) :
            LogManager().error(f'param is not the same [params.{self._parameter}]')
            return
            
        LogManager().debug('box parameter changed.')

        self._semaphore_values = True

        self._widget_value_editor.setValue(new_value)

        self._semaphore_values = False


    def init_ui (self) -> None :
        # Parameter Name Label
        self._parameter_name_label: PropertyLabel = PropertyLabel(self.parameter)
        
        # Parameter Value
        parameter_to_edit = self.box_model.get_parameter(self.parameter)
        self._widget_value_editor = property_widget_factory(parameter_to_edit)
        self._widget_value_editor.valueChanged.connect(self.on_value_changed)

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._parameter_name_label, row, 0)
        self.layout().addWidget(self._widget_value_editor, row, 1)
        # self.layout().addRow(self._parameter_name_label, QLabel("test"))

    def translate_ui (self) -> None:
        pass
        # self._lbl_parameter_name.setText(QCoreApplication.translate(self.objectName(), u"Nom : ", None))
        # self._lbl_parameter_type.setText(QCoreApplication.translate(self.objectName(), u"Type : ", None))
        # self._lbl_parameter_value.setText(QCoreApplication.translate(self.objectName(), u"Valeur : ", None))

    def unload (self) -> None :
        self._box_model.param_changed.disconnect(self.box_param_changed)


    def on_value_changed(self) -> None:
        if self._semaphore_values :
            return

        # Get old & new values
        new_value = self._widget_value_editor.value
        old_value = self.box_model.get_parameter(self.parameter)['value']
        param_name = 'params.' + self.parameter

        # Prepare action holder
        self.actions_generator(old_value, new_value, param_name)
        
        # Apply parameter value change
        self.box_model.action_param_semaphore = True
        self.box_model.set_parameter(self.parameter, new_value)
        self.box_model.action_param_semaphore = False


#-----------------------------------
# Imports
#-----------------------------------

from typing import Dict

#-----------------------------------
# Class
#-----------------------------------

from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel


class AbstractImporter :
    """Est la classe abstraite capable d'importer des modèles."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        ...

    # -------------
    # Methods
    # -------------

    def open (self, data : Dict, path : str, name : str) -> AbstractEditableModel :
        """Est la méthode appelée pour ouvrir le modèle éditable."""
        raise Exception("AbstractExporter open is an abstract method and must be overloaded.")


from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QApplication, QMainWindow
import sys


class MyWindow (QMainWindow) :
    def __init__ (self) :
        super(MyWindow, self).__init__()
        self.initUI()

    def initUI(self) :
        self.setGeometry(200, 200, 300, 300)
        self.setWindowTitle("PLCI")

        self._lbl_title = QtWidgets.QLabel(self)
        self._lbl_title.setText("PLCI")
        self._lbl_title.move(50, 50)

        self._btn_b1 = QtWidgets.QPushButton(self)
        self._btn_b1.setText("Button")
        self._btn_b1.clicked.connect(self.btn_b1_clicked)

    def btn_b1_clicked (self) :
        self._lbl_title.setText("Hello world lorem ipsum data Annalisa")
        self._lbl_title.adjustSize()
        print("Clicked !!")


def window() :
    app = QApplication(sys.argv)
    win = MyWindow()

    win.show()
    sys.exit(app.exec_())

window()



#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from .plci_core_data import Data
from .plci_core_data_type import DataType

#-----------------------------------
# Class
#-----------------------------------

class Signal :
    """Est la classe qui gère les différents signaux du diagramme."""

    # -------------
    # Constructors
    # -------------
    def __init__ (self, index_ : int, signal_type_ : DataType, default_value_ : Any, 
            name_: str = "", input_name: str = "", output_name: str = "") :

        self._index = index_
        self._signal_type = signal_type_
        self._data = Data(self._signal_type, default_value_)
        self._name = name_
        self._input_name = input_name
        self._output_name = output_name

    # -------------
    # Properties
    # -------------
    
    @property
    def index (self) -> int :
        """Permet de récuperer l'index."""    
        return self._index

    @property
    def signal_type (self) -> DataType :
        """Permet de récuperer le type de signal."""
        return self._signal_type
    
    @property
    def data (self) -> Data :
        """Permet de récuperer la donnée actuelle du signal."""
        return self._data

    @data.setter
    def data (self, data_ : Data) -> None : 
        """Permet de modifier la donnée interne du signal."""

        if (self._data.data_type == data_.data_type)  :
            self._data = data_
            return

    @property 
    def value (self) -> Any :
        """Permet de récuperer la valeur actuelle du signal."""
        return self._data.value

    @property 
    def name (self) -> str :
        """Permet de récuperer le nom du signal"""
        return self._name

    @name.setter
    def name (self, name_ : str):
        """Permet de modifier le nom du signal"""
        self._name = name_

    @property 
    def input_name (self) -> str :
        """Permet de récuperer le nom du signal"""
        return self._input_name

    @input_name.setter
    def input_name (self, new_input_name : str):
        """Permet de modifier le nom du signal"""
        self._input_name = new_input_name

    @property 
    def output_name (self) -> str :
        """Permet de récuperer le nom du signal"""
        return self._output_name

    @output_name.setter
    def output_name (self, new_output_name : str):
        """Permet de modifier le nom du signal"""
        self._output_name = new_output_name
        
    def __str__ (self) -> str :
        """Permet de retourner un string décrivant le signal."""
        return "Signal [ " + str(self.index) + " -> " + str(self._signal_type) + "]"


#-----------------------------------
# Imports
#-----------------------------------

# General Imports

# PyQt5 Imports
from PyQt5.QtCore import QCoreApplication, QPointF, QSizeF, Qt, pyqtSlot
from PyQt5.QtWidgets import QLabel

# Project Imports
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.properties_pages.rotation_slider import RotationSlider
from muphyn.packages.interface.models.graphical_models.abstract_moveable_graphical_element import AbstractMoveableGraphicalElement
from muphyn.packages.interface.models.dbl_spin_box import DblSpinBox
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel, TitlePropertyLabel

#-----------------------------------
# Class
#-----------------------------------

class MoveableGraphicalElementPropertiesEditor (AbstractPropertiesEditor) :
    """Est la classe permettant de modifier les paramètres d'un élément mobile dans la scène."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, moveable_element : AbstractMoveableGraphicalElement) :

        AbstractPropertiesEditor.__init__(self, moveable_element)

        self._size_semaphore = True
        self._pos_semaphore = True
        self._rot_semaphore = True

        self._moveable_element : AbstractMoveableGraphicalElement = None
        self.moveable_element = moveable_element

        self._size_semaphore = False
        self._pos_semaphore = False
        self._rot_semaphore = False

    
    # -------------
    # Properties
    # -------------

    @property
    def moveable_element (self) -> AbstractMoveableGraphicalElement :
        """Permet de récuperer l'élément graphique mobile."""
        return self._moveable_element

    @moveable_element.setter
    def moveable_element (self, moveable_element_ : AbstractMoveableGraphicalElement) -> None :
        """Permet de modifier l'élément graphique mobile."""
        
        if not(self._moveable_element is None) :

            self._moveable_element.size_changed.disconnect(self.element_size_changed)
            self._moveable_element.position_changed.disconnect(self.element_position_changed)
            self._moveable_element.rotation_changed.disconnect(self.element_rotation_changed)

        self._moveable_element = moveable_element_
        self._model = moveable_element_

        if not(self._moveable_element is None) :

            self._spn_x.setValue(self._moveable_element.x())
            self._spn_y.setValue(self._moveable_element.y())
            self._spn_width.setValue(self.moveable_element.size.width())
            self._spn_height.setValue(self.moveable_element.size.height())
            self._sldr_rotation.setValue(int(self._moveable_element.rotation()))
            self._lbl_rotation_value.setText(str(self._moveable_element.rotation()))

            self._moveable_element.size_changed.connect(self.element_size_changed)
            self._moveable_element.position_changed.connect(self.element_position_changed)
            self._moveable_element.rotation_changed.connect(self.element_rotation_changed)

    # -------------
    # Methods
    # -------------

    @pyqtSlot()
    def spn_pos_value_changed (self) -> None : 
        """Est la méthode appelée lorsque l'utilisateur change la position via une des deux spin box."""

        if self._pos_semaphore :
            return

        self._pos_semaphore = True

        self._moveable_element.position = QPointF(self._spn_x.value(), self._spn_y.value())

        self._pos_semaphore = False

    def element_size_changed (self) :
        """Est la méthode appelée lorsque la box change de taille."""
        
        if self._size_semaphore :
            return

        self._size_semaphore = True

        self._spn_width.setValue(self._moveable_element.size.width())
        self._spn_height.setValue(self._moveable_element.size.height())

        self._size_semaphore = False

    @pyqtSlot()
    def spn_size_value_changed (self) :
        """Est la méthode appelée lorsque l'utilisateur change la taille via une des deux spin box."""
    
        if self._pos_semaphore :
            return

        self._pos_semaphore = True

        # Get size
        new_size = QSizeF(self._spn_width.value(), self._spn_height.value())

        if new_size != self._moveable_element.size:

            # Set size
            self._moveable_element.size = QSizeF(self._spn_width.value(), self._spn_height.value())

            # Get size from moveable item
            self._spn_width.setValue(self._moveable_element.size.width())
            self._spn_height.setValue(self._moveable_element.size.height())

        self._pos_semaphore = False

    def element_position_changed (self) :
        """Est la méthode appelée lorsque la box change de taille."""
    
        if self._pos_semaphore :
            return

        self._pos_semaphore = True

        self._spn_x.setValue(self._moveable_element.position.x())
        self._spn_y.setValue(self._moveable_element.position.y())

        self._pos_semaphore = False

    def element_rotation_changed (self) :
        """Est la méthode appelée lorsque la box change de taille."""
        
        if self._rot_semaphore :
            return

        self._rot_semaphore = True

        self._sldr_rotation.setValue(self._moveable_element.rotation())
        self._lbl_rotation_value.setText(str(self._sldr_rotation.value()))

        self._rot_semaphore = False
        
    @pyqtSlot()
    def sldr_rotation_value_changed (self) -> None :
        """Est la méthode appelée lorsque l'utilisateur change la rotation de l'élément via le slider."""
        
        if self._rot_semaphore :
            return

        self._rot_semaphore = True

        self._lbl_rotation_value.setText(str(self._sldr_rotation.value()))
        self._moveable_element.setRotation(self._sldr_rotation.value())

        self._rot_semaphore = False 

    def init_ui (self) :
        
        if not self.objectName() :
            self.setObjectName(u"pnl_moveable_graphical_element_properties_editor")

        self._lbl_geometry : TitlePropertyLabel = TitlePropertyLabel()
        self._lbl_geometry.setObjectName(u"_lbl_geometry")

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_geometry, row, 0)

        self._lbl_x : PropertyLabel = PropertyLabel()
        self._lbl_x.setObjectName(u"_lbl_x")

        self._spn_x : DblSpinBox = DblSpinBox()
        self._spn_x.setObjectName(u"_spn_x")
        self._spn_x.setDecimals(0)
        self._spn_x.valueChanged.connect(self.spn_pos_value_changed)

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_x, row, 0)
        self.layout().addWidget(self._spn_x, row, 1)
        
        self._lbl_y : PropertyLabel = PropertyLabel()
        self._lbl_y.setObjectName(u"_lbl_y")

        self._spn_y : DblSpinBox = DblSpinBox()
        self._spn_y.setObjectName(u"_spn_y")
        self._spn_y.setDecimals(0)
        self._spn_y.valueChanged.connect(self.spn_pos_value_changed)

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_y, row, 0)
        self.layout().addWidget(self._spn_y, row, 1)

        self._lbl_width : PropertyLabel = PropertyLabel()
        self._lbl_width.setObjectName(u"_lbl_width")
        
        self._spn_width : DblSpinBox = DblSpinBox()
        self._spn_width.setObjectName(u"_spn_width")
        self._spn_width.setDecimals(0)
        self._spn_width.setMinimum(0)
        self._spn_width.valueChanged.connect(self.spn_size_value_changed)

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_width, row, 0)
        self.layout().addWidget(self._spn_width, row, 1)

        self._lbl_height : PropertyLabel = PropertyLabel()
        self._lbl_height.setObjectName(u"_lbl_y")
        
        self._spn_height : DblSpinBox = DblSpinBox()
        self._spn_height.setObjectName(u"_spn_height")
        self._spn_height.setDecimals(0)
        self._spn_height.valueChanged.connect(self.spn_size_value_changed)

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_height, row, 0)
        self.layout().addWidget(self._spn_height, row, 1)

        self._lbl_rotation : PropertyLabel = PropertyLabel()
        self._lbl_rotation.setObjectName(u"_lbl_rotation")

        self._sldr_rotation : RotationSlider = RotationSlider()
        self._sldr_rotation.setOrientation(Qt.Horizontal)
        self._sldr_rotation.setObjectName(u"_sldr_rotation")
        self._sldr_rotation.valueChanged.connect(self.sldr_rotation_value_changed)

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_rotation, row, 0)
        self.layout().addWidget(self._sldr_rotation, row, 1)

        self._lbl_rotation_value : PropertyLabel = PropertyLabel()
        self._lbl_rotation_value.setObjectName(u"_lbl_rotation_value")

        # Add Row
        row = self.layout().rowCount()
        self.layout().addWidget(self._lbl_rotation_value, row, 1)
    
    def translate_ui (self) -> None :
        self._lbl_geometry.setText(QCoreApplication.translate(self.objectName(), u"Geometry : ", None))
        self._lbl_x.setText(QCoreApplication.translate(self.objectName(), u"X : ", None))
        self._lbl_width.setText(QCoreApplication.translate(self.objectName(), u"Width : ", None))
        self._lbl_y.setText(QCoreApplication.translate(self.objectName(), u"Y : ", None))
        self._lbl_height.setText(QCoreApplication.translate(self.objectName(), u"Height : ", None))
        self._lbl_rotation.setText(QCoreApplication.translate(self.objectName(), u"Rotate angle : ", None))


#-----------------------------------
# Imports
#-----------------------------------

# PyQt5 imports
from typing import Any
from PyQt5.QtCore import QCoreApplication 
from PyQt5.QtWidgets import QWidget, QLabel, QLineEdit, QHBoxLayout

# Project imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.signals_model.abstract_connection_model import AbstractConnectionModel
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel

#-----------------------------------
# Class
#-----------------------------------

class InifiniteIOPropertiesItem (QWidget) :
    """Est l'item qui permet d'afficher des éléments visuels pour modifier une entrée/sortie dans la liste des entrées/sorites infinies."""

    # -------------
    # Constants
    # -------------
    ItemHeight: int = 40

    # -------------
    # Constructors
    # -------------

    def __init__ (self, number : int, connection_model : AbstractConnectionModel, parent : QWidget = None) :

        QWidget.__init__(self, parent)

        self._connection_model = connection_model

        self.init_ui()
        self.translate_ui()

        # Set field text
        self._fld_text.setText(connection_model.text)
        self.number = number

        # Connect text change
        self.connection_model.param_changed.connect(self.on_text_changed)

    # -------------
    # Methods
    # -------------

    @property
    def number (self) -> int :
        """Permet de récuperer le nombre de l'item."""
        return self._number

    @number.setter
    def number (self, number_ : int) -> None :
        """Permet de modifier le nombre de l'item."""
        self._number = number_
        self._lbl_index.setText(str(self._number) + ' : ')

    @property
    def connection_model (self) -> AbstractConnectionModel :
        """Permet de récuperer le modèle de la connexion."""
        return self._connection_model

    # -------------
    # Methods
    # -------------

    def init_ui (self) -> None :
        """Permet de créer et dessinier l'interface graphique."""

        self._lyt_ui : QHBoxLayout = QHBoxLayout()

        self._lbl_index : PropertyLabel = PropertyLabel()

        self._fld_text : QLineEdit = QLineEdit()
        self._fld_text.editingFinished.connect(self.text_field_edited)

        self._lyt_ui.addWidget(self._lbl_index)
        self._lyt_ui.addWidget(self._fld_text)

        self.setLayout(self._lyt_ui)

    def translate_ui (self) -> None :
        """Permet de traduire les éléments de l'interface graphique."""

        self._fld_text.setPlaceholderText(QCoreApplication.translate(self.objectName(), u"Texte", None))

    def text_field_edited (self) -> None : 
        """Est la méthode appelée lorsque l'utilisateur termine de modifier le champ d'étition du texte de l'entrée."""
        LogManager().debug(f'text field edited {self.number} new text : {self._fld_text.text()}')
        self._connection_model.text = self._fld_text.text()
        self._connection_model.name = self._fld_text.text()

    def on_text_changed(self, connection_model: AbstractConnectionModel, param_name: str, old_value: Any, new_value: Any) -> None:
        if param_name == "text":
            self._fld_text.setText(new_value)


#-----------------------------------
# Imports
#-----------------------------------

#-----------------------------------
# Class
#-----------------------------------

class AbstractGraphicalAction :
    """Est une la classe abstraite permettant de faire et défaire des actions graphiques."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        ...

    # -------------
    # Methods
    # -------------

    def do (self) :
        """Permet de réaliser l'action."""
        raise('AbstractGraphicalAction.do mathod is abstract and must be overhidden.')

    def undo (self) :
        """Permet de défaire l'action."""
        raise('AbstractGraphicalAction.undo mathod is abstract and must be overhidden.')


#-----------------------------------
# Imports
#-----------------------------------
import os
from typing import List
from muphyn.packages.core.plci_core_box import Box
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.core.simulation.om_sim import OpenModelicaSimulation
from muphyn.packages.core.simulation.model_parser.open_modelica_model_parser import OpenModelicaModelParser
from muphyn.packages.core.plci_core_scheduler_event import SchedulerEvent


"""
Notes:
 - If the initialization of the node is made in the _init_function then we won't access to the inputs first value 
    → (must implement a way to create firsts values)
 - If the outputs are read before the step is made we get one  step lag ;
 - If the inputs are set before the step then we get one step lag
"""

#-----------------------------------
# Functions
#-----------------------------------
def _step_simumation (box: Box, event_: SchedulerEvent) -> List :
    pass

def _finish_simulation (box: Box) -> List :
    for name, value in box.params.items():
        print(name, value)
    return []


#-----------------------------------
# Imports
#-----------------------------------

from typing import Iterable, List
from PyQt5 import QtGui
from PyQt5.QtCore import QCoreApplication, QRect, Qt
from PyQt5.QtWidgets import QLabel, QLineEdit, QWidget, QGridLayout
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibrariesManager
from muphyn.packages.interface.box_library.library_treeview import LibraryTreeView
from muphyn.packages.interface.widgets.BaseWidgets.Label import PropertyLabel

#-----------------------------------
# Class
#-----------------------------------

class BoxLibrariesList (QWidget) :
    """Est l'élément graphique qui permet de maintenir les bibliothèques de boxes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, parent : QWidget = None) :
        QWidget.__init__(self, parent)
        self.init_gui()

    # -------------
    # Properties
    # -------------

    @property
    def boxes_libraries (self) :
        """Permet de récuperer la bibliothèque de boxes."""
        return BoxesLibrariesManager()

    # -------------
    # Methods
    # -------------

    def searched_boxes (self) -> Iterable[AbstractBoxData] : 
        """Permet de récuperer les éléments après recherche."""

        search_text = self.fld_search.text() 
        elements : List[AbstractBoxData] = []

        if search_text.__len__() == 0 :
            for box_data in BoxesLibrariesManager().boxes :
                elements.append(box_data)

        else :
            keywords = search_text.split(' ')
            for box_data in BoxesLibrariesManager().boxes :

                b = True
                for k in keywords :
                    keyword = k.strip()

                    if keyword.__len__() == 0 :
                        continue
                    
                    if not(box_data._full_box_path.__contains__(keyword)) :
                        b = False
                        break

                if b : 
                    elements.append(box_data)

        elements.sort()
        
        for el in elements :
            yield el

    def resizeEvent (self, event : QtGui.QResizeEvent) -> None :

        self.tree_view.setGeometry(QRect(5, 30, self.width() - 15, self.height() - 35))
        self.fld_search.setGeometry(QRect(90, 1, self.width() - 100, 22))

        return super().resizeEvent(event)

    def libraries_reloaded (self) -> None :
        """Est la méthode appelée pour recharger la bibliothèque."""
        self.tree_view.build_model(BoxesLibrariesManager().boxes)
        
    def init_gui (self) -> None :
        """Est la méthode appelée pour afficher les éléments visuels dans la fenêtre."""
        
        if not self.objectName():
            self.setObjectName(u"box_libraries")

        # Init Layout
        layout = QGridLayout()
        layout.setContentsMargins(8, 8, 8, 8)
        layout.setColumnStretch(1, 1)
        layout.setRowStretch(1, 1)
        layout.setVerticalSpacing(10)
            
        # Init Tree View
        self.tree_view: LibraryTreeView = LibraryTreeView()
        self.tree_view.build_model(BoxesLibrariesManager().boxes)
        self.tree_view.setObjectName(u"lbr_treeview")
        self.tree_view.setMinimumHeight(250)

        # Init Search Field
        self.fld_search : QLineEdit = QLineEdit()
        self.fld_search.setObjectName(u"fld_search")
        self.fld_search.textChanged.connect(self.search_text_changed)

        # Init Search Label
        self.lbl_search : PropertyLabel = PropertyLabel()
        self.lbl_search.setObjectName(u"lbl_search")

        # Add widget in layout
        layout.addWidget(self.lbl_search, 0, 0, alignment=Qt.AlignmentFlag.AlignVCenter)
        layout.addWidget(self.fld_search, 0, 1, alignment=Qt.AlignmentFlag.AlignVCenter)
        layout.addWidget(self.tree_view, 1, 0, 1, 2)

        # Set Layout
        self.setLayout(layout)

        self.retranslateUi()
        
    def retranslateUi (self) -> None :
        self.lbl_search.setText(QCoreApplication.translate("Search : ", u"Search : ", None))

    def search_text_changed (self, searched_string: str) -> None :
        """Est la méthode appelée lorsque le texte contenus dans la barre de recherche est changée."""
        self.tree_view.filter_library_list(searched_string)



from typing import Optional, Union

from PyQt5.QtCore import Qt, QEvent, QRect
from PyQt5.QtGui import QCursor
from PyQt5.QtWidgets import QDialog, QWidget, QApplication

from muphyn.packages.core.simulation.model_parser.open_modelica_model_parser import OpenModelicaModelParser
from muphyn.packages.interface.editors.multiphysics_model.open_modelica_multiphysics_model import OpenModelicaMultiphysicsModel

class MultiphysicsModelPreviewer(QDialog):

    # Constants
    DefaultRect = QRect(0, 0, 640, 480)

    def __init__(self, multiphysics_parser: OpenModelicaModelParser, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent, Qt.WindowType.Dialog)
        
        # Get screen 
        globalCursorPosition = QCursor.pos()
        screenNumber = QApplication.desktop().screenNumber(globalCursorPosition)

        # Get current screen 
        screenRect = QApplication.desktop().availableGeometry(screenNumber)
        appRect = MultiphysicsModelPreviewer.DefaultRect

        # Calculate centered geometry
        offset = screenRect.center() - appRect.center()
        appRect.translate(offset)

        # General Parameters
        self.setMinimumSize(appRect.size())
        self.setGeometry(appRect)

        self.openModelicaModelPreviewer = OpenModelicaMultiphysicsModel(multiphysics_parser, parent=self)

    def event(self, a0: QEvent) -> bool:
        if a0.type() == QEvent.Type.Resize:
            self.openModelicaModelPreviewer.setSize(self.size())
        return super().event(a0)


#-----------------------------------
# Imports
#-----------------------------------

from typing import Dict
import importlib
import yaml

from muphyn.packages.core.scheduler_library.scheduler_library_data import SchedulerData
from muphyn.packages.core.utils.log_manager import LogManager
from .abstract_scheduler_library_importer import AbstractSchedulerLibraryImporter

#-----------------------------------
# Class
#-----------------------------------

class SchedulerLibraryImporter_1_0_0 (AbstractSchedulerLibraryImporter) : 
    """Est la V 1.0.0 de classes capable d'importer des planificateurs."""

    # -------------
    # Constructors
    # -------------
    
    def __init__ (self) :
        super().__init__()

    # -------------
    # Methods
    # -------------

    def import_scheduler_data (self, path : str, file_name : str, absolute_yaml_file : str, schedulers : Dict[str, SchedulerData]) -> Dict[str, SchedulerData] :
        
        absolute_py_file = path + '/' + file_name + '.py'

        with open(absolute_yaml_file) as yaml_file_data :
            
            file_data = yaml.load(yaml_file_data, Loader = yaml.FullLoader)

            if 'scheduler' in file_data:
                
                scheduler_data = file_data['scheduler']

                if scheduler_data['library'].startswith('Schedulers') :
                    
                    library_name = scheduler_data['library'] + "." + scheduler_data['name']
                    spec = importlib.util.spec_from_file_location(library_name, absolute_py_file)
                    foo = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(foo)

                    if scheduler_data['scheduler_method'] == "None" : 
                        scheduler_method = lambda box : LogManager().error("No scheduler method")
                    else :
                        scheduler_method = getattr(foo, scheduler_data['scheduler_method'])

                    return {'library_name' : library_name, 'scheduler_data' : SchedulerData(
                        scheduler_name = scheduler_data['name'],
                        scheduler_library = scheduler_data['library'],
                        scheduler_method = scheduler_method,
                        creator = scheduler_data['creator'],
                        date_created = scheduler_data['date_creation'],
                        version = scheduler_data['version']
                    )}


#-----------------------------------
# Imports
#-----------------------------------
# General Imports
from typing import Iterable, Any

# PyQt5 Imports
from PyQt5 import QtGui
from PyQt5.QtCore import QMarginsF, QPointF, QRectF, pyqtSlot
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QGraphicsItem

# Project Imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.links_model.link_type import LinkType
from muphyn.packages.interface.shapes.basic_shape.shapes import Path
from muphyn.packages.interface.utils.interface_constants import MuphynPens

#-----------------------------------
# Class
#-----------------------------------
class AbstractLinkModel (AbstractGraphicalElement) :
    """Est la classe abstraites aux classes représentant des liens entre les noeuds."""

    # -------------
    # Constants
    # -------------
    LINK_TO_BOX_MARGIN = 20
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, input : Any, output : Any, link_type : LinkType, 
                  text : str = '') :
        
        # Save Input & Output model
        self._input_model : Any = input
        self._output_model : Any = output

        # Get absolute position of Input & Output
        input_position : QPointF = self._input_model.absolute_connector_center
        output_position : QPointF = self._output_model.absolute_connector_center

        # Init previous position
        self._last_input_position = input_position
        self._last_output_position = output_position

        # Points
        points = AbstractLinkModel.calc_path_points(input_position, output_position)

        # Calculate origin
        bounding_rect = AbstractLinkModel.calc_bounding_rect(points)

        # Calculate all points from origin
        points = [point - bounding_rect.topLeft() for point in points]
        
        # Calculate steps
        if link_type == LinkType.SQUARE:
            steps = AbstractLinkModel.build_straight_path(points)
        elif link_type == LinkType.CURVED:
            steps = AbstractLinkModel.build_curved_path(points)
        else:
            LogManager().Error("AbstractLinkModel.__init__() : Wrong link type {type(link_type)}")
            return

        # Set Position
        AbstractGraphicalElement.__init__(self, 'link', bounding_rect.topLeft(), 0, text, None)
        
        # Append path
        self._path = Path(steps, parent=self)
        self._path.pen = MuphynPens.SelectedLinkPen if self.isSelected() else MuphynPens.UnSelectedLinkPen

        # Set Link parameters
        self._link_type : LinkType = link_type
        self._link_value : float = 0.8

        if self._input_model.parent() is not None :
            self._input_model.parent().position_changed.connect(self.input_changed)
            self._input_model.parent().size_changed.connect(self.input_changed)
            self._input_model.parent().rotation_changed.connect(self.input_changed)

        if self._output_model.parent() is not None:
            self._output_model.parent().position_changed.connect(self.output_changed)
            self._output_model.parent().size_changed.connect(self.output_changed)
            self._output_model.parent().rotation_changed.connect(self.output_changed)

        
        self._bounding_rect = bounding_rect.translated(-bounding_rect.topLeft())

        self._about_to_erased = False

    # -------------
    # Properties
    # -------------
    @AbstractGraphicalElement.rot.setter
    def rot (self, rot_ : float) -> None :
        ...

    @property
    def color(self) -> QColor:
        return self._path.border_color
    
    @property
    def input (self) -> Any :
        """Permet de récuperer le noeud d'entrée du lien."""
        return self._input_model

    @property 
    def output (self) -> Any :
        """Permet de récuperer le noeud de sortie du lien."""
        return self._output_model

    @property
    def link_type (self) -> LinkType :
        """Permet de récuperer le type de lien entre les deux noeuds."""
        return self._link_type

    @link_type.setter
    def link_type (self, link_type_ : LinkType) -> None :
        """Permet de modifier le type de lien entre les deux noeuds."""

        if self._link_type == link_type_ :
            return

        old_value = self._link_type

        self._link_type = link_type_

        if self._link_value > 1.0 or self._link_value < 0.0 :
            self._link_value = 0.5

        if self.action_param_semaphore :
            return

        self.param_changed.emit(self, 'link_type', old_value, self._link_type)

    @property
    def link_value (self) -> float :
        """
        Permet de récuperer la valeur du lien :
        - en cas de lien courbe : la courbure du lien.
        - en cas de lien carré : le pourcentage avant que le lien ne se brise.
        """
        return self._link_value

    @link_value.setter
    def link_value (self, link_value_ : float) -> None :
        """Permet de modifier la valeur du lien."""

        if self._link_value == link_value_ :
            return

        old_value = self._link_value

        self._link_value = link_value_

        if self._link_value > 1.0 or self._link_value < 0.0 :
            self._link_value = 0.5

        if self.action_param_semaphore :
            return

        self.param_changed.emit(self, 'link_value', old_value, self._link_value)


    # -------------
    # Methods
    # -------------
    @pyqtSlot()
    def input_changed (self) :
        """Est la méthode appelée lorsque la poistion du noeud 1 change."""
        self.self_update()

    @pyqtSlot()
    def output_changed (self) :
        """Est la méthode appelée lorsque la position du noeud 2 change."""
        self.self_update()

    def boundingRect(self) -> QRectF:
        return self._bounding_rect

    def setRotation(self, rot_ : float) -> None :
        ...
 
    def self_update (self) -> None :
        """Est la méthode appelée lorsque la position ou la taille du node 2 est modifiée."""
        if self.scene() is None :
            return

        last_bounding_rect = self._path.boundingRect().marginsAdded(QMarginsF(5, 5, 5, 5))

        # get new input & output positions
        input_position = self._input_model.absolute_connector_center
        output_position = self._output_model.absolute_connector_center
        
        # Points
        points = AbstractLinkModel.calc_path_points(input_position, output_position)

        # Calculate origin
        min_x = min([point.x() for point in points])
        min_y = min([point.y() for point in points])
        top_left = QPointF(min_x, min_y)

        # Calculate all points from origin
        points = [point - top_left for point in points]
        
        # Calculate steps
        steps = AbstractLinkModel.build_straight_path(points)
        self._path.steps = steps

        # Update Link Position
        self.setPos(top_left)

        # Calc bounding rect
        min_x = min(input_position.x(), self._last_input_position.x(), output_position.x(), self._last_output_position.x())
        min_y = min(input_position.y(), self._last_input_position.y(), output_position.y(), self._last_output_position.y())
        max_x = max(input_position.x(), self._last_input_position.x(), output_position.x(), self._last_output_position.x())
        max_y = max(input_position.y(), self._last_input_position.y(), output_position.y(), self._last_output_position.y())
        width = abs(max_x - min_x)
        height = abs(max_y - min_y)

        
        new_bounding_rect = self._path.boundingRect().marginsAdded(QMarginsF(5, 5, 5, 5))

        # Update Scene 
        self.scene().update(min_x, min_y, new_bounding_rect.width(), new_bounding_rect.height())

        self._bounding_rect = new_bounding_rect

        self._last_input_position = input_position
        self._last_output_position = output_position

    def unbind (self) -> None :
        """Permet de détruire le lien entre les deux noeuds."""

        if self._about_to_erased :
            return

        if self in self._input_model._links :
            self._input_model.remove_link(self)
            self._input_model.update(self._input_model.boundingRect())

        if self in self._output_model._links :
            self._output_model.remove_link(self)
            self._output_model.update(self._output_model.boundingRect())
        
        self._about_to_erased = True

    def itemChange (self, change: QGraphicsItem.GraphicsItemChange, value: Any) -> Any:
        if change == QGraphicsItem.GraphicsItemChange.ItemSelectedHasChanged :
            self._is_selected = bool(value)
            self._path.pen = MuphynPens.SelectedLinkPen if self._is_selected else MuphynPens.UnSelectedLinkPen
            self.self_update()

        return super().itemChange(change, value)

    # --------------
    # Static Methods
    # --------------
    def calc_path_points(start_point: QPointF, end_point: QPointF):
        """
        This function assumes that the link start from the input of a box
        & ends in the output of the other box
        """
        # 2 points case
        """
        end_point[]--x
                     |
                     x--[] start_point
        """
        # Set Margin to start/end points
        start_point.setX(start_point.x() - AbstractLinkModel.LINK_TO_BOX_MARGIN)
        end_point.setX(end_point.x() + AbstractLinkModel.LINK_TO_BOX_MARGIN)

        # Round points position
        start_point = QPointF(start_point.toPoint())
        end_point = QPointF(end_point.toPoint())

        if start_point.x() > end_point.x():
            if start_point.y() == end_point.y():
                return [
                    start_point + QPointF(AbstractLinkModel.LINK_TO_BOX_MARGIN, 0),
                    end_point - QPointF(AbstractLinkModel.LINK_TO_BOX_MARGIN, 0)
                ]
            else:
                # Calculate intermediate X
                x_intermediate = (start_point.x() + end_point.x()) / 2

                # Append Intermediate points
                return [
                    start_point + QPointF(AbstractLinkModel.LINK_TO_BOX_MARGIN, 0),
                    QPointF(x_intermediate, start_point.y()),
                    QPointF(x_intermediate, end_point.y()),
                    end_point - QPointF(AbstractLinkModel.LINK_TO_BOX_MARGIN, 0)
                ]
        else:
            # 4 points case
            """
            x-[] 
            |
            x---------x
                      |     
                   []-x
            """
            # Calculate intermediate coordinates
            y_intermediate = int((start_point.y() + end_point.y()) / 2)

            # Append Intermediate points
            return [
                start_point + QPointF(AbstractLinkModel.LINK_TO_BOX_MARGIN, 0),
                start_point,
                QPointF(start_point.x(), y_intermediate),
                QPointF(end_point.x(), y_intermediate),
                end_point,
                end_point - QPointF(AbstractLinkModel.LINK_TO_BOX_MARGIN, 0),
            ]

    def calc_bounding_rect(points: Iterable[QPointF]) -> QRectF:
        min_x = min([point.x() for point in points])
        max_x = max([point.x() for point in points])
        min_y = min([point.y() for point in points])
        max_y = max([point.y() for point in points])

        width = abs(max_x - min_x)
        height = abs(max_y - min_y)
        return QRectF(min_x, min_y, width, height)

    def build_straight_path(points: Iterable[QPointF]):
        return [Path.Step(point) for point in points]

    def build_curved_path(points: Iterable[QPointF]):
        factor = 0.2
        path = QtGui.QPainterPath(points[0])

        last_end_point = QPointF
        for index, corner in enumerate(points[1: -1]):
            # Calculate middle of corner segments
            middle_point_1 = (points[index] + corner) / 2
            middle_point_2 = (points[index+2] + corner) / 2

            start_point = (1-factor) * middle_point_1 + factor * corner
            end_point = (1-factor) * middle_point_2 + factor * corner

            path.cubicTo(start_point, corner, end_point)

        path.lineTo(points[-1])

        return path


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

#-----------------------------------
# Functions
#-----------------------------------

def _function_multiplier_box (box, event_) -> List:
    v = 1

    for input in box.inputs:
        v *= input.value

    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events